from __future__ import absolute_import, print_function, unicode_literals

from typing import Dict, Optional
from doozerlib.distgit import pull_image
from doozerlib.metadata import Metadata
from doozerlib.model import Missing, Model
import hashlib
import pathlib
import os
import json
import re
from functools import lru_cache

from doozerlib.pushd import Dir
from doozerlib import assertion, exectools
from dockerfile_parse import DockerfileParser
from doozerlib import brew
from doozerlib.util import BlackBox


class ImageMetadata(Metadata):

    def __init__(self, runtime, data_obj: Dict, commitish: Optional[str] = None):
        super(ImageMetadata, self).__init__('image', runtime, data_obj, commitish)
        self.image_name = self.config.name
        self.required = self.config.get('required', False)
        self.image_name_short = self.image_name.split('/')[-1]
        self.parent = None
        self.children = []  # list of ImageMetadata which use this image as a parent.
        dependents = self.config.get('dependents', [])
        for d in dependents:
            self.children.append(self.runtime.late_resolve_image(d, add=True))

    def is_ancestor(self, image):
        """
        :param image: A Metadata object or a distgit_key to check.
        :return: Returns whether the specified image is an ancestor of this image (eg. a parent, or parent's parent...)
        """
        if isinstance(image, Metadata):
            image = image.distgit_key

        parent = self.parent
        while parent:
            if parent.distgit_key == image:
                return True
            parent = parent.parent

        return False

    def get_descendants(self):
        """
        :return: Returns a set of children imagemetadata and their children, ..etc.
        """
        descendants = set()
        for child in self.children:
            descendants.add(child)
            descendants.update(child.get_descendants())

        return descendants

    def resolve_parent(self):
        """
        :return: Resolves and returns imagemeta.parent attribute for this image's parent image OR None if no parent is defined.
        """
        if 'from' in self.config:
            if 'member' in self.config['from']:
                base = self.config['from']['member']
                try:
                    self.parent = self.runtime.resolve_image(base)
                except:
                    self.parent = None

                if self.parent:
                    self.parent.add_child(self)
        return self.parent

    def add_child(self, child):
        """
        Adds a child imagemetadata to this list of children for this image.
        :param child:
        :return:
        """
        self.children.append(child)

    @property
    def base_only(self):
        """
        Some images are marked base-only.  Return the flag from the config file
        if present.
        """
        return self.config.base_only

    @property
    def is_payload(self):
        return self.config.get('for_payload', False)

    def get_component_name(self, default=-1):
        """
        :param default: Not used. Here to stay consistent with similar rpmcfg.get_component_name
        :return: Returns the component name of the image. This is the name in the nvr
        that brew assigns to this image's build.
        """
        # By default, the bugzilla component is the name of the distgit,
        # but this can be overridden in the config yaml.
        component_name = self.name

        # For apbs, component name seems to have -apb appended.
        # ex. http://dist-git.host.prod.eng.bos.redhat.com/cgit/apbs/openshift-enterprise-mediawiki/tree/Dockerfile?h=rhaos-3.7-rhel-7
        if self.namespace == "apbs":
            component_name = "%s-apb" % component_name

        if self.namespace == "containers":
            component_name = "%s-container" % component_name

        if self.config.distgit.component is not Missing:
            component_name = self.config.distgit.component

        return component_name

    def get_brew_image_name_short(self):
        # Get image name in the Brew pullspec. e.g. openshift3/ose-ansible --> openshift3-ose-ansible
        return self.image_name.replace("/", "-")

    def pull_url(self):
        # Don't trust what is the Dockerfile for version & release. This field may not even be present.
        # Query brew to find the most recently built release for this component version.
        _, version, release = self.get_latest_build_info()

        # we need to pull images from proxy since https://mojo.redhat.com/docs/DOC-1204856 if 'brew_image_namespace'
        # is enabled.
        if self.runtime.group_config.urls.brew_image_namespace is not Missing:
            name = self.runtime.group_config.urls.brew_image_namespace + '/' + self.config.name.replace('/', '-')
        else:
            name = self.config.name

        return "{host}/{name}:{version}-{release}".format(
            host=self.runtime.group_config.urls.brew_image_host, name=name, version=version,
            release=release)

    def pull_image(self):
        pull_image(self.pull_url())

    def get_default_push_tags(self, version, release):
        push_tags = [
            "%s-%s" % (version, release),  # e.g. "v3.7.0-0.114.0.0"
            "%s" % version,  # e.g. "v3.7.0"
        ]

        # it's possible but rare that an image will have an alternate
        # tags along with the regular ones
        # append those to the tag list.
        if self.config.push.additional_tags is not Missing:
            push_tags.extend(self.config.push.additional_tags)

        # In v3.7, we use the last .0 in the release as a bump field to differentiate
        # image refreshes. Strip this off since OCP will have no knowledge of it when reaching
        # out for its node image.
        if "." in release:
            # Strip off the last field; "0.114.0.0" -> "0.114.0"
            push_tags.append("%s-%s" % (version, release.rsplit(".", 1)[0]))

        # Push as v3.X; "v3.7.0" -> "v3.7"
        push_tags.append("%s" % (version.rsplit(".", 1)[0]))
        return push_tags

    def get_default_repos(self):
        """
        :return: Returns a list of ['ns/repo', 'ns/repo'] found in the image config yaml specified for default pushes.
        """
        # Repos default to just the name of the image (e.g. 'openshift3/node')
        default_repos = [self.config.name]

        # Unless overridden in the config.yml
        if self.config.push.repos is not Missing:
            default_repos = self.config.push.repos.primitive()

        return default_repos

    def get_default_push_names(self):
        """
        :return: Returns a list of push names that should be pushed to for registries defined in
        group.yml and for additional repos defined in image config yaml.
        (e.g. ['registry/ns/repo', 'registry/ns/repo', ...]).
        """

        # Will be built to include a list of 'registry/ns/repo'
        push_names = []

        default_repos = self.get_default_repos()  # Get a list of [ ns/repo, ns/repo, ...]

        default_registries = []
        if self.runtime.group_config.push.registries is not Missing:
            default_registries = self.runtime.group_config.push.registries.primitive()

        for registry in default_registries:
            registry = registry.rstrip("/")   # Remove any trailing slash to avoid mistaking it for a namespace
            for repo in default_repos:
                namespace, repo_name = repo.split('/')
                if '/' in registry:  # If registry overrides namespace
                    registry, namespace = registry.split('/')
                push_names.append('{}/{}/{}'.format(registry, namespace, repo_name))

        # image config can contain fully qualified image names to push to (registry/ns/repo)
        if self.config.push.also is not Missing:
            push_names.extend(self.config.push.also)

        return push_names

    def get_additional_push_names(self, additional_registries):
        """
        :return: Returns a list of push names based on a list of additional registries that
        need to be pushed to (e.g. ['registry/ns/repo', 'registry/ns/repo', ...]).
        """

        if not additional_registries:
            return []

        # Will be built to include a list of 'registry/ns/repo'
        push_names = []

        default_repos = self.get_default_repos()  # Get a list of [ ns/repo, ns/repo, ...]

        for registry in additional_registries:
            registry = registry.rstrip("/")   # Remove any trailing slash to avoid mistaking it for a namespace
            for repo in default_repos:
                namespace, repo_name = repo.split('/')
                if '/' in registry:  # If registry overrides namespace
                    registry, namespace = registry.split('/')
                push_names.append('{}/{}/{}'.format(registry, namespace, repo_name))

        return push_names

    # Class methods to speed up computation of does_image_need_change if multiple
    # images will be assessed.
    # Mapping of brew pullspec => most recent brew build dict.
    builder_image_builds = dict()

    def does_image_need_change(self, eldest_image_event_ts, changing_rpm_packages=[], buildroot_tag_ids=None):
        """
        TODO: this method is threadsafe now. Call it with threads!
        Answers the question of whether the latest built image needs to be rebuilt based on
        the packages (and therefore RPMs) it is dependent on might have changed in tags
        relevant to the image. A check is also made if the image depends on a package
        we know is changing because we are about to rebuild it.
        :param eldest_image_event_ts: The brew timestamp for the creation of the oldest image in this group.
        :param changing_rpm_packages: A list of package names that are about to change.
        :param buildroot_tag_ids: A list of build tag id's that contribute to this image's build root.
        :return: (<bool>, messsage). If True, the image might need to be rebuilt -- the message will say
                why. If False, message will be None.
        """

        dgk = self.distgit_key
        runtime = self.runtime

        rpm_entries_sets = {}  # Each key is an archive image id. Each value is a query of that archive's RPMs.
        with runtime.pooled_koji_client_session() as koji_api:

            image_build = self.get_latest_build(default='')
            if not image_build:
                # Seems this have never been built. Mark it as needing change.
                return True, 'Image has never been built before'

            self.logger.debug(f'Image {dgk} latest is {image_build}')

            image_nvr = image_build['nvr']
            image_build_event_id = image_build['creation_event_id']  # the brew event that created this build

            self.logger.info(f'Running an change assessment on {image_nvr} built at event {image_build_event_id}')

            # Very rarely, an image might need to pull an package that is not actually installed in the
            # builder image or in the final image.
            # e.g. https://github.com/openshift/ironic-ipa-downloader/blob/999c80f17472d5dbbd4775d901e1be026b239652/Dockerfile.ocp#L11-L14
            # This is programmatically undetectable through koji queries. So we allow extra scan-sources hints to
            # be placed in the image metadata.
            if self.config.scan_sources.extra_packages is not Missing:
                for package_details in self.config.scan_sources.extra_packages:
                    extra_package_name = package_details.name
                    extra_package_brew_tag = package_details.tag
                    # Example output: https://gist.github.com/jupierce/3bbc8be7265348a8f549d401664c9972
                    extra_latest_tagging_infos = koji_api.tagHistory(package=extra_package_name, tag=extra_package_brew_tag, queryOpts={'limit': 1})
                    if not extra_latest_tagging_infos:
                        self.logger.warning(f'Unable to find tagging event for for extra_packages {extra_package_name}')
                        continue
                    # Otherwise, we have information about the most recent time this package was tagged into the
                    # relevant tag. Why the tagging event and not the build time? Well, the build could have been
                    # made long ago, but only tagged into the relevant tag recently.
                    extra_latest_tagging_event = extra_latest_tagging_infos[0]['create_event']
                    self.logger.debug(f'Checking image creation time against extra_packages {extra_package_name} in tag {extra_package_brew_tag} @ tagging event {extra_latest_tagging_event}')
                    if extra_latest_tagging_event > image_build_event_id:
                        return True, f'Image {dgk} is sensitive to extra_packages {extra_package_name} which changed at event {extra_latest_tagging_event}'

            # Collect build times from any builder images used to create this image
            builders = self.config['from'].builder or []
            for builder in builders:
                if builder.member:
                    # We can't determine if images are about to change. Defer to scan-sources.
                    continue

                if builder.image:
                    builder_image_name = builder.image
                elif builder.stream:
                    builder_image_name = runtime.resolve_stream(builder.stream).image
                else:
                    raise IOError(f'Unable to determine builder image pullspec from {builder}')

                # builder_image_name example: "openshift/ose-base:ubi8"
                brew_image_url = self.runtime.resolve_brew_image_url(builder_image_name)
                builder_brew_build = ImageMetadata.builder_image_builds.get(brew_image_url, None)

                if not builder_brew_build:
                    out, err = exectools.cmd_assert(f'oc image info {brew_image_url} --filter-by-os amd64 -o=json', retries=5, pollrate=10)
                    latest_builder_image_info = Model(json.loads(out, encoding='utf-8'))
                    builder_info_labels = latest_builder_image_info.config.config.Labels
                    builder_nvr_list = [builder_info_labels['com.redhat.component'], builder_info_labels['version'], builder_info_labels['release']]

                    if not all(builder_nvr_list):
                        raise IOError(f'Unable to find nvr in {builder_info_labels}')

                    builder_image_nvr = '-'.join(builder_nvr_list)
                    builder_brew_build = koji_api.getBuild(builder_image_nvr)
                    ImageMetadata.builder_image_builds[brew_image_url] = builder_brew_build
                    self.logger.debug(f'Found that builder image {brew_image_url} has event {builder_brew_build}')

                if image_build_event_id < builder_brew_build['creation_event_id']:
                    self.logger.info(f'will be rebuilt because a builder image changed: {builder_image_name}')
                    return True, f'A builder image {builder_image_name} has changed since {image_nvr} was built'

            build_root_changes = brew.tags_changed_since_build(runtime, koji_api, image_build, buildroot_tag_ids)
            if build_root_changes:
                changing_tag_names = [brc['tag_name'] for brc in build_root_changes]
                self.logger.info(f'Image will be rebuilt due to buildroot change since {image_nvr} (last build event={image_build_event_id}). Build root changes changes: [{changing_tag_names}]')
                self.logger.debug(f'Image will be rebuilt due to buildroot change since ({image_build}) (last build event={image_build_event_id}). Build root changes: {build_root_changes}')
                return True, f'Buildroot tag changes in [{changing_tag_names}] since {image_nvr}'

            archives = koji_api.listArchives(image_build['id'])

            # Compare to the arches in runtime
            build_arches = {a['extra']['image']['arch'] for a in archives}
            target_arches = set(self.get_arches())
            if target_arches != build_arches:
                # The latest brew build does not exactly match the required arches as specified in group.yml
                return True, f'Arches of {image_nvr}: ({build_arches}) does not match target arches {target_arches}'

            for archive in archives:
                # Example results of listing RPMs in an given imageID:
                # https://gist.github.com/jupierce/a8798858104dcf6dfa4bd1d6dd99d2d8
                archive_id = archive['id']
                rpm_entries = koji_api.listRPMs(imageID=archive_id)
                rpm_entries_sets[archive_id] = rpm_entries

        # Leave the pooled koji context manager, so the following method can't deadlock waiting for an api.
        # The method we call inside also uses the pool.
        for archive_id, rpm_entries in rpm_entries_sets.items():
            self.logger.info(f'Checking whether one of my latest archives ({archive_id}) has any of its {len(rpm_entries)} rpms associated with a package that has been tagged in relevant tag since this image\'s build brew event {image_build_event_id}')
            rpm_nvrs = [entry['nvr'] for entry in rpm_entries]
            self.logger.debug(f'RPMS from {archive_id}: {rpm_nvrs}')
            changes_res = runtime.parallel_exec(
                f=lambda idx, terminate_event: is_image_older_than_package_build_tagging(self, image_build_event_id, idx, rpm_entries, changing_rpm_packages, eldest_image_event_ts),
                args=range(len(rpm_entries)),
                n_threads=20
            )

            for changed, msg in changes_res.get():
                if changed:
                    return True, msg

        return False, None

    def covscan(self, result_archive, repo_type='unsigned', local_repo=[]):
        self.logger.info('Setting up for coverity scan')
        all_js = 'all_results.js'
        diff_js = 'diff_results.js'
        all_html = 'all_results.html'
        diff_html = 'diff_results.html'
        waived_flag = 'waived.flag'

        archive_path = pathlib.Path(result_archive)
        dg_archive_path = archive_path.joinpath(self.distgit_key)
        dg_archive_path.mkdir(parents=True, exist_ok=True)  # /<archive-dir>/<dg-key>

        builders = self.config['from'].builder
        if builders is Missing:
            self.logger.info('No builder images -- does not appear to be container first. Skipping.')
            return

        dgr = self.distgit_repo()
        with Dir(dgr.distgit_dir):
            dg_commit_hash, _ = exectools.cmd_assert('git rev-parse HEAD', strip=True)
            archive_commit_results_path = dg_archive_path.joinpath(dg_commit_hash)  # /<archive-dir>/<dg-key>/<hash>
            archive_all_results_js_path = archive_commit_results_path.joinpath(all_js)  # /<archive-dir>/<dg-key>/<hash>/all_results.js
            archive_all_results_html_path = archive_commit_results_path.joinpath(all_html)  # /<archive-dir>/<dg-key>/<hash>/all_results.html
            archive_diff_results_js_path = archive_commit_results_path.joinpath(diff_js)
            archive_diff_results_html_path = archive_commit_results_path.joinpath(diff_html)
            archive_waived_flag_path = archive_commit_results_path.joinpath(waived_flag)

            def write_record():
                diff = json.loads(archive_diff_results_js_path.read_text(encoding='utf-8'))
                diff_count = len(diff['issues'])
                if diff_count == 0:
                    self.logger.info('No new issues found during scan')
                    archive_waived_flag_path.write_text('')  # No new differences, mark as waived

                owners = ",".join(self.config.owners or [])
                self.runtime.add_record('covscan',
                                        distgit=self.qualified_name,
                                        distgit_key=self.distgit_key,
                                        commit_results_path=str(archive_commit_results_path),
                                        all_results_js_path=str(archive_all_results_js_path),
                                        all_results_html_path=str(archive_all_results_html_path),
                                        diff_results_js_path=str(archive_diff_results_js_path),
                                        diff_results_html_path=str(archive_diff_results_html_path),
                                        diff_count=str(diff_count),
                                        waive_path=str(archive_waived_flag_path),
                                        waived=str(archive_waived_flag_path.exists()).lower(),
                                        owners=owners,
                                        image=self.config.name,
                                        commit_hash=dg_commit_hash)

            if archive_diff_results_html_path.exists():
                self.logger.info(f'This commit already has scan results ({str(archive_all_results_js_path)}; skipping scan')
                write_record()
                return

            archive_commit_results_path.mkdir(parents=True, exist_ok=True)

            dg_path = pathlib.Path(Dir.getcwd())
            cov_path = dg_path.joinpath('cov')

            dockerfile_path = dg_path.joinpath('Dockerfile')
            if not dockerfile_path.exists():
                self.logger.error('Dockerfile does not exist in distgit; not rebased yet?')
                return

            dfp = DockerfileParser(str(dockerfile_path))
            covscan_builder_df_path = dg_path.joinpath('Dockerfile.covscan.builder')

            with covscan_builder_df_path.open(mode='w+') as df_out:
                first_parent = dfp.parent_images[0]
                ns, repo_tag = first_parent.split(
                    '/')  # e.g. openshift/golang-builder:latest => [openshift, golang-builder:latest]
                if '@' in repo_tag:
                    repo, tag = repo_tag.split(
                        '@')  # e.g. golang-builder@sha256:12345 =>  golang-builder & tag=sha256:12345
                    tag = '@' + tag
                else:
                    if ':' in repo_tag:
                        repo, tag = repo_tag.split(':')  # e.g. golang-builder:latest =>  golang-builder & tag=latest
                    else:
                        repo = repo_tag
                        tag = 'latest'
                    tag = ':' + tag

                first_parent_url = f'registry-proxy.engineering.redhat.com/rh-osbs/{ns}-{repo}{tag}'

                # build a local image name we can use.
                # We will build a local scanner image based on the target distgit's first parent.
                # It will have coverity tools installed.
                m = hashlib.md5()
                m.update(first_parent.encode('utf-8'))
                local_builder_tag = f'{repo}-{m.hexdigest()}'

                vol_mount_arg = ''
                make_image_repo_files = ''

                if local_repo:
                    for idx, lr in enumerate(local_repo):
                        make_image_repo_files += f"""
# Create a repo able to pull from the local filesystem and prioritize it for speed.
RUN echo '[covscan_local_{idx}]' > /etc/yum.repos.d/covscan_local_{idx}.repo
RUN echo 'baseurl=file:///covscan_local_{idx}' >> /etc/yum.repos.d/covscan_local_{idx}.repo
RUN echo skip_if_unavailable=True >> /etc/yum.repos.d/covscan_local_{idx}.repo
RUN echo gpgcheck=0 >> /etc/yum.repos.d/covscan_local_{idx}.repo
RUN echo enabled=1 >> /etc/yum.repos.d/covscan_local_{idx}.repo
RUN echo enabled_metadata=1 >> /etc/yum.repos.d/covscan_local_{idx}.repo
RUN echo priority=1 >> /etc/yum.repos.d/covscan_local_{idx}.repo
    """
                        vol_mount_arg += f' -mount {lr}:/covscan_local_{idx}'
                else:
                    make_image_repo_files = 'RUN wget https://cov01.lab.eng.brq.redhat.com/coverity/install/covscan/covscan-rhel-7.repo -O /etc/yum.repos.d/covscan.repo\n'

                df_out.write(f'''FROM {first_parent_url}
LABEL DOOZER_COVSCAN_PARENT_IMAGE=true
LABEL DOOZER_COVSCAN_FIRST_PARENT={local_builder_tag}
LABEL DOOZER_COVSCAN_GROUP={self.runtime.group_config.name}

{make_image_repo_files}

RUN yum install -y wget

RUN wget {self.cgit_url(".oit/" + repo_type + ".repo")} -O /etc/yum.repos.d/oit.repo
RUN yum install -y python36

# Enable epel for csmock
RUN wget https://dl.fedoraproject.org/pub/epel/epel-release-latest-7.noarch.rpm
RUN yum -y install epel-release-latest-7.noarch.rpm

# Certs necessary to install from covscan repos
RUN wget https://password.corp.redhat.com/RH-IT-Root-CA.crt -O /etc/pki/ca-trust/source/anchors/RH-IT-Root-CA.crt --no-check-certificate
RUN wget https://password.corp.redhat.com/legacy.crt -O /etc/pki/ca-trust/source/anchors/legacy.crt --no-check-certificate
RUN wget https://engineering.redhat.com/Eng-CA.crt -O /etc/pki/ca-trust/source/anchors/Eng-CA.crt --no-check-certificate
RUN update-ca-trust
RUN update-ca-trust enable

RUN yum install -y cov-sa csmock csmock-plugin-coverity csdiff
    ''')

            rc, out, err = exectools.cmd_gather(f'docker image inspect {local_builder_tag}')
            if rc != 0:
                rc, _, _ = exectools.cmd_gather(f'imagebuilder {vol_mount_arg} -t {local_builder_tag} -f {str(covscan_builder_df_path)} {str(dg_path)}')
                if rc != 0:
                    self.logger.error(f'Unable to create scanner image based on builder image: {first_parent_url}')
                    # TODO: log this as a record and make sure the pipeline warns artist.
                    # until covscan can be installed on rhel-8, this is expected for some images.
                    return

            # We should now have an image tagged local_builder_tag based on the original image's builder image
            # but also with covscan tools installed.

            covscan_df_path = dg_path.joinpath('Dockerfile.covscan')

            runs = ['#!/bin/bash', 'set -o xtrace']
            setup = []
            first_from = False
            for entry in dfp.structure:
                content = entry['content']
                instruction = entry['instruction'].upper()
                if instruction.upper() == 'FROM':
                    if first_from:
                        break
                    first_from = True

                if instruction == 'COPY':
                    setup.append(content)

                if instruction == 'ENV':
                    setup.append(content)

                if instruction == 'RUN':
                    runs.append(content.strip()[4:])

                if instruction == 'WORKDIR':
                    setup.append(content)  # Pass into setup so things like ADD work as expected
                    path = content.strip()[7:]
                    runs.append('mkdir -p ' + path)  # Also pass into RUN so they have the correct working dir
                    runs.append('cd ' + path)

            run_file = '\n'.join(runs)
            self.logger.info(f'Constructed run file:\n{run_file}')

            build_script_name = 'doozer_covscan_runit.sh'
            with dg_path.joinpath(build_script_name).open(mode='w+') as runit:
                runit.write(run_file)

            with covscan_df_path.open(mode='w+') as df_out:
                df_out.write(f'FROM {local_builder_tag}\n')
                df_out.write(f'LABEL DOOZER_COVSCAN_GROUP={self.runtime.group_config.name}\n')
                df_out.write(f'ADD {build_script_name} /\n')
                df_out.write(f'RUN chmod +x /{build_script_name}\n')
                df_out.write('\n'.join(setup) + '\n')
                df_out.write('ENV PATH=/opt/coverity/bin:${PATH}\n')

            run_tag = f'{local_builder_tag}.{self.image_name_short}'
            rc, out, err = exectools.cmd_gather(f'docker image inspect {run_tag}')
            if rc != 0:
                exectools.cmd_assert(f'imagebuilder -t {run_tag} -f {str(covscan_df_path)} {str(dg_path)}')

            cov_path.mkdir(exist_ok=True)
            emit_path = cov_path.joinpath('emit')
            if not emit_path.exists():
                rc, out, err = exectools.cmd_gather(f'docker run --hostname=covscan --rm -v {str(cov_path)}:/cov:z {run_tag} cov-build --dir=/cov /{build_script_name}')
                if rc != 0:
                    self.logger.error('Did not achieve full compilation')

                builg_log_path = cov_path.joinpath('build-log.txt')
                build_log = builg_log_path.read_text(encoding='utf-8')
                if '[WARNING] No files were emitted' in build_log:
                    self.logger.error(f'Build did not emit anything. Check out the build-log.txt: {builg_log_path}')
                    # TODO: log this as a record and make sure the pipeline warns artist
                    return

            else:
                self.logger.info('covscan emit already exists -- skipping this step')

            def run_docker_cov(cmd):
                return exectools.cmd_assert(f'docker run --hostname=covscan --rm -v {str(cov_path)}:/cov:z {run_tag} {cmd}')

            summary_path = cov_path.joinpath('output', 'summary.txt')
            if not summary_path.exists() or 'Time taken by analysis' not in summary_path.read_text(encoding='utf-8'):
                # This can take an extremely long time and use virtually all CPU
                run_docker_cov('cov-analyze  --dir=/cov "--wait-for-license" "-co" "ASSERT_SIDE_EFFECT:macro_name_lacks:^assert_(return|se)\\$" "-co" "BAD_FREE:allow_first_field:true" "--include-java" "--fb-max-mem=4096" "--security" "--concurrency" --allow-unmerged-emits')
            else:
                self.logger.info('covscan analysis already exists -- skipping this step')

            all_results_js, _ = run_docker_cov('cov-format-errors --json-output-v2 /dev/stdout --dir=/cov')
            all_results_js_path = cov_path.joinpath(all_js)
            all_results_js_path.write_text(all_results_js, encoding='utf-8')

            all_results_html = "<html>Error generating HTML report.</html>"
            try:
                # Rarely, cshtml just outputs empty html and rc==1; just ignore it.
                all_results_html, _ = run_docker_cov(f'cshtml /cov/{all_js}')
            except:
                self.logger.warning(f'Error generating HTML report for {str(archive_all_results_js_path)}')
                pass

            all_results_html_path = cov_path.joinpath(all_html)
            all_results_html_path.write_text(all_results_html, encoding='utf-8')

            run_docker_cov(f'chown -R {os.getuid()}:{os.getgid()} /cov')  # Otherwise, root will own these files

            # Write out the files to the archive directory as well
            archive_all_results_js_path.write_text(all_results_js, encoding='utf-8')
            archive_all_results_html_path.write_text(all_results_html, encoding='utf-8')

            # Now on to computing diffs

            # Search backwards through commit history; try to find a hash for this distgit that has been scanned before
            diff_results_js = all_results_js  # Unless we find an old has, diff results matches all results
            commit_log, _ = exectools.cmd_assert("git --no-pager log --pretty='%H' -1000")
            for old_commit in commit_log.split()[1:]:
                old_all_results_js_path = dg_archive_path.joinpath(old_commit, all_js)
                old_are_results_waived_path = dg_archive_path.joinpath(old_commit, waived_flag)
                # Only compute diff from commit if results were actually waived
                # This file should be created by the Jenkins / scanning pipeline.
                if old_are_results_waived_path.exists():
                    diff_results_js, _ = exectools.cmd_assert(f'csdiff {str(archive_all_results_js_path)} {str(old_all_results_js_path)}')
                    break

            archive_diff_results_js_path.write_text(diff_results_js, encoding='utf-8')

            diff_results_html = "<html>Error generating HTML report.</html>"
            try:
                # Rarely, cshtml just outputs empty html and rc==1; just ignore it.
                diff_results_html, _ = exectools.cmd_assert(f'cshtml {str(archive_diff_results_js_path)}')
            except:
                self.logger.warning(f'Error generating HTML report for {str(archive_diff_results_js_path)}')
                pass

            archive_diff_results_html_path.write_text(diff_results_html, encoding='utf-8')

            write_record()


@lru_cache(maxsize=None)  # cache without bound
def _get_tag_info(koji_api_black_box: BlackBox, tag_name):
    return koji_api_black_box.contents.getTag(tag_name)


def get_tag_info(koji_api, tag_name):
    # Returns (a potentially cached) copy of this tag's information.
    # See examples: https://gist.github.com/jupierce/adeb7b2b10f5d225c8090bab80640011
    return _get_tag_info(BlackBox(koji_api), tag_name)


@lru_cache(maxsize=None)  # cache without bound
def _get_tag_history(koji_api_black_box: BlackBox, limit=None, **kwargs):
    return koji_api_black_box.contents.tagHistory(queryOpts={'limit': limit}, **kwargs)


def get_tag_history(koji_api, limit=None, **kwargs):
    return _get_tag_history(BlackBox(koji_api), limit=limit, **kwargs)


@lru_cache(maxsize=None)  # cache without bound
def _get_build(koji_api_black_box: BlackBox, *args, **kwargs):
    return koji_api_black_box.contents.getBuild(*args, **kwargs)


def get_build(koji_api, *args, **kwargs):
    return _get_build(BlackBox(koji_api), *args, **kwargs)


@lru_cache(maxsize=None)  # cache without bound
def _query_history(koji_api_black_box: BlackBox, table, *args, **kwargs):
    return koji_api_black_box.contents.queryHistory([table], *args, **kwargs)


def query_history(koji_api, table, *args, **kwargs):
    return _query_history(BlackBox(koji_api), table, *args, **kwargs)


def is_image_older_than_package_build_tagging(image_meta, image_build_event_id, ri, rpm_entries, changing_rpm_packages, eldest_image_event_ts):
    """
    Determines if a given rpm is part of a package that has been tagged in a relevant tag AFTER image_build_event_id
    :param image_meta: The image meta object for the image
    :param image_build_event_id: The build event for the image
    :param ri: An index into the list of rpm_entries to check
    :param rpm_entries: A set of rpms to analyze (return of koji_api.listRPMs)
    :param changing_rpm_packages: A list of package name that the caller knows to be changing in the future
    :param eldest_image_event_ts: The brew timestamp for the creation of the oldest image in this group.
    :return: (<bool>, message) . If True, the message will describe the change reason. If False, message will
            will be None.
    """

    # If you are considering changing this code, you are going to have to contend with
    # complex scenarios like: what if we pulled in this RPM by tagging it, intentionally
    # backlevel, from another product, and we want to keep like that? Or, what if we tagged
    # in a non-released version of another product to get a fix pre-release of that package,
    # but subsequently want to inherit later versions the original product ships.
    # This blunt approach isn't trying to be perfect, but it will rarely do an unnecessary
    # rebuild and handles those complex scenarios by erring on the side of doing the rebuild.

    runtime = image_meta.runtime

    with runtime.pooled_koji_client_session() as koji_api:
        rpm_entry = rpm_entries[ri]
        package_build_id = rpm_entry['build_id']

        # Find all tags ever applied to this build. The theory goes is that one of these tags is
        # the reason why ultimately pulled this particular package build into our image.
        # Example output: https://gist.github.com/jupierce/dd84a024711e5a6e33f0fc7842c12151
        build_tag_history = get_tag_history(koji_api, build=package_build_id)
        build_tag_set = set()
        for bt_hist_entry in build_tag_history:
            build_tag_set.add(bt_hist_entry['tag_name'])

        package_name = build_tag_history[0]['name']  # Extract the package name from one of these tagging events
        if package_name in changing_rpm_packages:
            image_meta.logger.info(f'Image must ust change because of forthcoming RPM package build {package_name}')
            return True, f'Image depends on package {package_name}, which was identified as about to change'

        # Now, collect a list of all tagging events that have applied to this PACKAGE since the oldest
        # image of this group was built. Without this limiting timestamp, all historical tagging events for a
        # package could number in the thousands.
        # This could be a BUNCH of tags that are irrelevant - tags for RHEL8 & RHEL7 / etc. To
        # winnow this list to something reasonable, we only care about tags we ALSO find in
        # the tags applied to the specific package build in our image (build_tag_history). Thus,
        # if our build is RHEL7, we won't accidentally detect a random RHEL8 tag as being relevant.
        # Example output: https://gist.github.com/jupierce/3b0e28f52fc12d117a37d6b7593ed608
        # This method caches (if all arguments are identical) and is thread safe.
        package_tag_history = query_history(koji_api, table='tag_listing', package=package_name, after=eldest_image_event_ts)['tag_listing']

        # There is a hole here. Given an RPM X with a history of tags {x}, we know we received & installed
        # the RPM through one of the tags in {x}. What happens if a new RPM Y had a set of tags {y} that
        # is fully disjoint from {x}. In short, it came in through a completely independent vector, but we
        # would still find it through that vector if a build was triggered.
        # This could happen if:
        # 1. group.yml repos are changed and Y would be pulled from a new source
        # 2. The new Y is available through a different repo than we found it in last time.
        # To mitigate #1, we should force build after changing group.yml repos.
        # For #2, the only way this would typically happen would be if we were pulling from an official
        # repo for rpm X and then Y was tagged into our candidate tag as an override. To account for
        # this, always check for changes in our tags.
        build_tag_set.add(image_meta.branch())
        build_tag_set.add(image_meta.candidate_brew_tag())

        relevant_tagging_events = {}  # tag_name => brew_event which may cause our next image build to pull in a new build of the package

        image_meta.logger.debug(f'Computed relevant tags for image rpm {rpm_entry["nvr"]}: {build_tag_set}')

        def register_relevant_tag_event(tag_name, event_id):
            nonlocal relevant_tagging_events
            # We only want the very latest events registered.
            if tag_name not in relevant_tagging_events:
                relevant_tagging_events[tag_name] = event_id
            else:
                if relevant_tagging_events[tag_name] < event_id:
                    relevant_tagging_events[tag_name] = event_id

        for tag_event in package_tag_history:
            tag_name = tag_event['tag.name']
            tagging_event = tag_event['create_event']

            if tag_name not in build_tag_set:
                # We exclude any tag that wasn't involved in how our existing image found the package
                continue

            if tag_name == 'trashcan' or '-private' in tag_name or 'deleted' in tag_name:
                # Exclude garbage tags
                continue

            if tag_name.endswith(('-released', '-set', '-pending', '-backup')):
                # Ignore errata tags (e.g. RHBA-2020:2309-released, RHBA-2020:3027-pending) and tags like rhel-8.0.0-z-batch-0.3-set
                continue

            if tag_name.startswith(image_meta.branch()):
                # If this was tagged with this image's brew_tag or brew_tag-candidate, it likely
                # can affect our next build.
                register_relevant_tag_event(tag_name, tagging_event)
                continue

            if tag_name.endswith(('-candidate', '-build')):
                # Now we can eliminate any other candidate tags and build tags
                continue

            if re.match(r".+-rhel-\d+$", tag_name):
                # Check for released tag naming convention. e.g. ends with rhel-7 or rhel-8.
                register_relevant_tag_event(tag_name, tagging_event)
                continue

            # Now for a fuzzier match assuming naming conventions are not followed.
            # We don't care about non-release builds from other products / oddball tags, so we can
            # filter out those tags. We do this by querying the tags themselves and seeing
            # if they are 'perm':'trusted'. Trusted tags are those
            # that seem to be related to shipped advisories.
            # See examples: https://gist.github.com/jupierce/adeb7b2b10f5d225c8090bab80640011
            tag_info = get_tag_info(koji_api, tag_name)   # This method is cached and thread safe.
            if not tag_info:
                # Some sort of pseudo tag like rhel-8.0.0-z-batch-0.3-set
                continue

            tag_perm = tag_info.get('perm', None)

            if not tag_perm:
                # Looks to be someone else's unconventional candidate tag or other non-shipment related tag.
                # rhaos-4.4-rhel-7 has tag_perm=='trusted'
                # kpatch-kernel-4.18.0-193.6.3.el8_2-build  has tag_perm='admin'
                continue

            # If we arrive here, register the tag as relevant
            register_relevant_tag_event(tag_name, tagging_event)

        for tag_name, rel_tagging_event_id in relevant_tagging_events.items():
            # tag_name now represents indicates another product shipped this build. Through various
            # methods, our image build could have picked it up from a yum repo associated with this tag.
            # We aren't going to try to figure it out for certain. Just see if our package changed
            # with respect to this relevant tag AFTER our image was built.
            if rel_tagging_event_id > image_build_event_id:
                # This tagging occurred after our image was built and was tagged with
                # a tag that may have affected the image's inputs. Rebuild to be
                # certain.
                msg = f'Possible rpm change because package "{package_name}" changed in tag "{tag_name}" at event: {rel_tagging_event_id}'
                image_meta.logger.info(msg)
                # For the humans reading the output, let's output the tag names that mattered.
                return True, msg

    return False, None



import json
import multiprocessing
import sys
from typing import Iterable, List, Optional, Tuple
from urllib.parse import urlparse

import click
import yaml

from doozerlib import Runtime, brew, embargo_detector, exectools
from doozerlib.cli import cli, pass_runtime
from doozerlib.exceptions import DoozerFatalError
from doozerlib.util import green_print


@click.group("detect-embargo", short_help="Check whether one or more images, RPMs, or release payloads have embargoed fixes.")
def detect_embargo():
    pass


cli.add_command(detect_embargo)


@detect_embargo.command("nvr", short_help="Detect embargoed fixes in given builds")
@click.option("--yaml", "as_yaml", is_flag=True,
              help="Print out the result as YAML format.")
@click.option("--json", "as_json", is_flag=True,
              help="Print out the result as JSON format.")
@click.argument("nvrs", metavar="NVRS...", nargs=-1, required=True)
@pass_runtime
def detect_in_nvr(runtime: Runtime, nvrs, as_yaml, as_json):
    """ Check whether one or more Brew builds have embargoed fixes.

    If neither --yaml nor --json is given, this program will exit with status 0 if true, or with status 2 if not.
    Errors are signaled by a non-zero status that is not 2.

    Example:

        $ doozer --group=openshift-4.6 detect-embargo nvr openshift-4.6.0-202007110420.p1.git.0.4de1d1d.el8 openshift-clients-4.6.0-202007152229.p0.git.3651.9a1d25f.el8
    """
    if as_yaml and as_json:
        raise click.BadParameter("Must use one of --yaml or --json.")
    runtime.initialize(clone_distgits=False)
    embargoed_builds = detect_embargoes_in_nvrs(runtime, nvrs)
    print_result_and_exit(embargoed_builds, None, None, as_yaml, as_json)


@detect_embargo.command("tag", short_help="Detect embargoed fixes in builds with given Koji tags")
@click.option("--kind", "kind", default="all", metavar="KIND", type=click.Choice(["rpm", "image", "all"]),
              help="Only detect the specified kind of builds. [rpm, image, all] (Default to all)")
@click.option("--exclude", "excluded_tags", metavar="TAG", multiple=True,
              help="Koji tag name that the build must not have [multiple]")
@click.option("--event-id", metavar='NUM', required=False, type=int,
              help="A Brew event ID. If specified, the last NVRs as of the given Brew event will be chosen instead of latest")
@click.option("--yaml", "as_yaml", is_flag=True,
              help="Print out the result as YAML format.")
@click.option("--json", "as_json", is_flag=True,
              help="Print out the result as JSON format.")
@click.argument("tags", metavar="TAGS...", nargs=-1, required=True)
@pass_runtime
def detect_in_tag(runtime: Runtime, kind, tags, excluded_tags, event_id, as_yaml, as_json):
    """ Check whether one or more Brew tags have builds that include embargoed fixes.

    If neither --yaml nor --json is given, this program will exit with status 0 if true, or with status 2 if not.
    Errors are signaled by a non-zero status that is not 2.

    Example:
        $ doozer --group=openshift-4.6 detect-embargo tag rhaos-4.6-rhel-8-candidate rhaos-4.6-rhel-7-candidate --exclude rhaos-4.6-rhel-8  --exclude rhaos-4.6-rhel-7
    """
    if as_yaml and as_json:
        raise click.BadParameter("Must use one of --yaml or --json.")
    runtime.initialize(clone_distgits=False)
    embargoed_builds = detect_embargoes_in_tags(runtime, kind, tags, excluded_tags, event_id)
    print_result_and_exit(embargoed_builds, None, None, as_yaml, as_json)


@detect_embargo.command("pullspec", short_help="Check whether one or more arbitrary brew images (referred by pullspecs) has embargoed fixes.")
@click.option("--yaml", "as_yaml", is_flag=True,
              help="Print out the result as YAML format.")
@click.option("--json", "as_json", is_flag=True,
              help="Print out the result as JSON format.")
@click.argument("pullspecs", metavar="PULLSPECS...", nargs=-1, required=True)
@pass_runtime
def detect_in_pullspec(runtime, pullspecs, as_yaml, as_json):
    """ Check whether one or more arbitrary container images (referred by pullspecs) have embargoed fixes.

    If neither --yaml nor --json is given, this program will exit with status 0 if true, or with status 2 if not.
    Errors are signaled by a non-zero status that is not 2.

    Example:

        $ doozer --group=openshift-4.6 detect-embargo pullspec registry-proxy.engineering.redhat.com/rh-osbs/openshift-ose-ironic@sha256:caa56dc45580ec5103564125cbb87f41559aec89de4a8efd099777463381180c
    """
    if as_yaml and as_json:
        raise click.BadParameter("Must use one of --yaml or --json.")
    runtime.initialize(clone_distgits=False)
    embargoed_pullspecs, embargoed_builds = detect_embargoes_in_pullspecs(runtime, pullspecs)
    print_result_and_exit(embargoed_builds, embargoed_pullspecs, None, as_yaml, as_json)


@detect_embargo.command("release", short_help="Check whether one or more release payloads has embargoed fixes.")
@click.option("--yaml", "as_yaml", is_flag=True,
              help="Print out the result as YAML format.")
@click.option("--json", "as_json", is_flag=True,
              help="Print out the result as JSON format.")
@click.argument("pullspecs", metavar="PULLSPECS...", nargs=-1, required=True)
@pass_runtime
def detect_in_release(runtime, pullspecs, as_yaml, as_json):
    """ Check whether one or more release payloads have images with embargoed fixes.

    If neither --yaml nor --json is given, this program will exit with status 0 if true, or with status 2 if not.
    Errors are signaled by a non-zero status that is not 2.

    Example:

        $ doozer --group=openshift-4.6 detect-embargo release registry.svc.ci.openshift.org/ocp/release:4.6.0-0.nightly-2020-07-17-043219
    """
    if as_yaml and as_json:
        raise click.BadParameter("Must use one of --yaml or --json.")
    runtime.initialize(clone_distgits=False)
    embargoed_releases, embargoed_pullspecs, embargoed_builds = detect_embargoes_in_releases(runtime, pullspecs)
    print_result_and_exit(embargoed_builds, embargoed_pullspecs, embargoed_releases, as_yaml, as_json)


def detect_embargoes_in_nvrs(runtime: Runtime, nvrs: List[str]):
    """ Finds embargoes in given NVRs
    :param runtime: the runtime
    :param nvrs: list of build NVRs
    :return: list of Brew build dicts that have embargoed fixes
    """
    runtime.logger.info(f"Fetching {len(nvrs)} builds from Brew...")
    brew_session = runtime.build_retrying_koji_client()
    builds = brew.get_build_objects(nvrs, brew_session)
    for i, b in enumerate(builds):
        if not b:
            raise DoozerFatalError(f"Unable to get {nvrs[i]} from Brew.")
    runtime.logger.info(f"Detecting embargoes for {len(nvrs)} builds...")
    detector = embargo_detector.EmbargoDetector(brew_session, runtime.logger)
    embargoed_build_ids = detector.find_embargoed_builds(builds)
    embargoed_builds = [b for b in builds if b["id"] in embargoed_build_ids]
    return embargoed_builds


def detect_embargoes_in_tags(runtime: Runtime, kind: str, included_tags: List[str], excluded_tags: List[str], event_id: Optional[int]):
    """ Finds embargoes in builds with given tags
    :param runtime: the runtime
    :param included_tags: list of koji tags that the returned builds must have
    :param excluded_tags: list of koji tags that the returned builds must not have
    :return: list of Brew build dicts that have embargoed fixes
    """
    brew_session = runtime.build_retrying_koji_client()
    runtime.logger.info(f"Fetching builds from Brew tags {included_tags}...")
    build_type = None if kind == "all" else kind
    latest_build_lists = brew.get_latest_builds([(tag, None) for tag in included_tags], build_type, event_id, brew_session)
    included_builds = [b for builds in latest_build_lists if builds for b in builds]  # flatten latest_build_lists
    runtime.logger.info(f"Found {len(included_builds)} builds from Brew tags {included_tags}.")
    if included_builds and excluded_tags:  # if we have tags to exclude, get all builds with excluded_tags then exclude them
        runtime.logger.info(f"Fetching builds from Brew tags {excluded_tags}...")
        excluded_build_lists = brew.get_tagged_builds(excluded_tags, build_type, event_id, brew_session)
        excluded_build_ids = {b["id"] for builds in excluded_build_lists if builds for b in builds}
        builds = [b for b in included_builds if b["id"] not in excluded_build_ids]
        runtime.logger.info(f"Excluded {len(included_builds) - len(builds)} builds that are also tagged into {excluded_tags}.")
        included_builds = builds

    # Builds may have duplicate entries if we query from multiple tags. Don't worry, EmbargoDetector is smart.
    runtime.logger.info(f"Detecting embargoes for {len(included_builds)} builds...")
    detector = embargo_detector.EmbargoDetector(brew_session, runtime.logger)
    embargoed_build_ids = detector.find_embargoed_builds(included_builds)
    embargoed_builds = [b for b in included_builds if b["id"] in embargoed_build_ids]
    return embargoed_builds


def detect_embargoes_in_pullspecs(runtime: Runtime, pullspecs: List[str]):
    """ Finds embargoes in given image pullspecs
    :param runtime: the runtime
    :param nvrs: list of image pullspecs
    :return: list of Brew build dicts that have embargoed fixes
    """
    runtime.logger.info(f"Fetching manifests for {len(pullspecs)} pullspecs...")
    jobs = runtime.parallel_exec(lambda pullspec, _: get_nvr_by_pullspec(pullspec), pullspecs,
                                 min(len(pullspecs), multiprocessing.cpu_count() * 4, 32))
    nvrs = jobs.get()
    suspect_nvrs = []
    suspect_pullspecs = []
    for index, nvr in enumerate(nvrs):
        n, v, r = nvr
        if not n or not v or not r:
            runtime.logger.warning(f"Assuming {pullspecs[index]} is not embargoed because it doesn't have valid NVR labels.")
            continue
        suspect_nvrs.append(f"{n}-{v}-{r}")
        suspect_pullspecs.append(pullspecs[index])

    embargoed_builds = detect_embargoes_in_nvrs(runtime, suspect_nvrs)
    embargoed_build_nvrs = {b["nvr"] for b in embargoed_builds}
    embargoed_pullspecs = [pullspec for index, pullspec in enumerate(suspect_pullspecs) if suspect_nvrs[index] in embargoed_build_nvrs]
    return embargoed_pullspecs, embargoed_builds


def detect_embargoes_in_releases(runtime: Runtime, pullspecs: List[str]):
    """ Finds embargoes in given release payloads
    :param runtime: the runtime
    :param nvrs: list of release pullspecs
    :return: list of Brew build dicts that have embargoed fixes
    """
    runtime.logger.info(f"Fetching component pullspecs from {len(pullspecs)} release payloads...")
    jobs = runtime.parallel_exec(lambda pullspec, _: get_image_pullspecs_from_release_payload(pullspec), pullspecs,
                                 min(len(pullspecs), multiprocessing.cpu_count() * 4, 32))
    pullspec_lists = jobs.get()
    embargoed_releases = []
    embargoed_pullspecs = []
    embargoed_builds = []
    for index, image_pullspecs in enumerate(pullspec_lists):
        p, b = detect_embargoes_in_pullspecs(runtime, list(image_pullspecs))
        if p:  # release has embargoes
            embargoed_releases.append(pullspecs[index])
            embargoed_pullspecs += p
            embargoed_builds += b
    return embargoed_releases, embargoed_pullspecs, embargoed_builds


def print_result_and_exit(embargoed_builds, embargoed_pullspecs, embargoed_releases, as_yaml, as_json):
    """ Prints embargo detection result and exit
    :param embargoed_builds: list of dicts of embargoed builds
    :embargoed_pullspecs: list of embargoed image pullspecs
    :embargoed_releases: list of pullspecs of embargoed release payloads
    :as_yaml: if true, print the result as an YAML document
    :as_json: if true, print the result as a JSON document
    """
    out = {
        "has_embargoes": bool(embargoed_builds)
    }
    if embargoed_builds:
        out["builds"] = embargoed_builds
    if embargoed_pullspecs:
        out["pullspecs"] = embargoed_pullspecs
    if embargoed_builds:
        out["releases"] = embargoed_releases
    if as_yaml:
        yaml.dump(out, sys.stdout)
    elif as_json:
        json.dump(out, sys.stdout)
    elif not embargoed_builds:
        green_print("No builds contain embargoed fixes.")
        exit(2)  # we use this special exit status to indicate no embargoed fixes are detected.
        return  # won't reach here. it is for spoofing the unittest when exit is mocked.
    else:
        if embargoed_releases:
            green_print(f"Found {len(embargoed_releases)} release payload containing embargoed fixes:")
            for release in embargoed_releases:
                green_print(release)
        green_print(f"Found {len(embargoed_builds)} builds containing embargoed fixes:")
        for index, build in enumerate(embargoed_builds):
            line = f"{build['id']}\t{build['nvr']}"
            if embargoed_pullspecs:
                line += f"\t{embargoed_pullspecs[index]}"
            green_print(line)
    exit(0)


def get_nvr_by_pullspec(pullspec: str) -> Tuple[str, str, str]:
    """ Retrieves (name, version, release) of the image referenced by a pullspec
    :param pullspec: image pullspec
    :return: a (name, version, release) tuple
    """
    if not urlparse(pullspec).scheme:
        pullspec = "docker://" + pullspec
    # If pullspec references a manifest list, 'skopeo inspect --config' should be able to process it under the hood
    out, _ = exectools.cmd_assert(["skopeo", "inspect", "--config", "--", pullspec])
    image_info = json.loads(out)
    labels = image_info["config"]["Labels"]
    return (labels.get("com.redhat.component"), labels.get("version"), labels.get("release"))


def get_image_pullspecs_from_release_payload(payload_pullspec: str, ignore={"machine-os-content"}) -> Iterable[str]:
    """ Retrieves pullspecs of images in a release payload.
    :param payload_pullspec: release payload pullspec
    :param ignore: a set of image names that we want to exclude from the return value (e.g. machine-os-content)
    :return: an iterator over pullspecs of component images
    """
    out, _ = exectools.cmd_assert(["oc", "adm", "release", "info", "-o", "json", "--", payload_pullspec])
    payload_info = json.loads(out)
    for tag in payload_info["references"]["spec"]["tags"]:
        if tag["name"] in ignore:
            continue
        yield tag["from"]["name"]



import glob
import io
import json
import os
import re
import threading
import yaml

from dockerfile_parse import DockerfileParser
from doozerlib import brew, exectools, logutil, pushd


class OLMBundle(object):
    """This class is responsible for generating bundle containers out of previously built operators

    Every OLM Operator image should have a corresponding bundle container, which is mostly empty,
    carrying only the operator's manifests and some special Dockerfile labels, that allows the
    bundle container to publish those manifests on operator's behalf

    Even though each bundle container has its dedicated distgit repo, they are not meant to be
    independently built, due to their tight coupling to corresponding operators
    """

    def __init__(self, runtime):
        self.runtime = runtime

    def rebase(self, operator_nvr):
        """Update bundle distgit contents with manifests from given operator NVR
        Perform image SHA replacement on manifests before commit & push
        Annotations and Dockerfile labels are re-generated with info from operator's package YAML

        :param string operator_nvr: Operator NVR (format: my-operator-v4.2.30-202004200449)
        :return bool True if rebase succeeds, False if there was nothing new to commit
        """
        self.operator_nvr = operator_nvr
        self.get_operator_buildinfo()
        self.clone_operator()
        self.checkout_operator_to_build_commit()
        self.clone_bundle()
        self.clean_bundle_contents()
        self.get_operator_package_yaml_info()
        self.copy_operator_manifests_to_bundle()
        self.replace_image_references_by_sha_on_bundle_manifests()
        self.generate_bundle_annotations()
        self.generate_bundle_dockerfile()
        self.create_container_yaml()
        return self.commit_and_push_bundle(commit_msg="Update bundle manifests")

    def build(self, operator_name=None):
        """Trigger a brew build of operator's bundle

        :param string operator_name: Operator name (as in ocp-build-data file name, not brew component)
        :return bool True if build succeeds, False otherwise
        """
        if operator_name:
            self.operator_repo_name = 'containers/{}'.format(operator_name)

        self.clone_bundle()

        if not self.trigger_bundle_container_build():
            return False

        return self.watch_bundle_container_build()

    def get_latest_bundle_build(self):
        """Get NVR of latest bundle build tagged on given target

        :return string: NVR of latest bundle build, or "" if there is none.
        """
        _rc, out, _err = exectools.cmd_gather(
            'brew latest-build --quiet {} {}'.format(self.target, self.bundle_brew_component)
        )
        return out.split(' ')[0]

    def get_operator_buildinfo(self):
        """Get operator distgit repository name and commit hash used to build given operator NVR
        """
        operator_buildinfo = brew.get_build_objects([self.operator_nvr], self.brew_session)[0]
        match = re.search(r'([^#]+)#(\w+)', operator_buildinfo['source'])

        self.operator_repo_name = '/'.join(match.group(1).split('/')[-2:])
        self.operator_build_commit = match.group(2)

    def clone_operator(self):
        """Clone operator distgit repository to doozer working dir
        """
        exectools.cmd_assert('rm -rf {}'.format(self.operator_clone_path))
        exectools.cmd_assert('mkdir -p {}'.format(os.path.dirname(self.operator_clone_path)))
        exectools.cmd_assert('rhpkg{}clone --branch {} {} {}'.format(
            self.rhpkg_opts, self.branch, self.operator_repo_name, self.operator_clone_path
        ), retries=3)

    def checkout_operator_to_build_commit(self):
        """Checkout clone of operator repository to specific commit used to build given operator NVR
        """
        with pushd.Dir(self.operator_clone_path):
            exectools.cmd_assert('git checkout {}'.format(self.operator_build_commit))

    def clone_bundle(self):
        """Clone corresponding bundle distgit repository of given operator NVR
        """
        exectools.cmd_assert('rm -rf {}'.format(self.bundle_clone_path))
        exectools.cmd_assert('mkdir -p {}'.format(os.path.dirname(self.bundle_clone_path)))
        exectools.cmd_assert('rhpkg{}clone --branch {} {} {}'.format(
            self.rhpkg_opts, self.branch, self.bundle_repo_name, self.bundle_clone_path
        ), retries=3)

    def clean_bundle_contents(self):
        """Delete all files currently present in the bundle repository
        Generating bundle files is an idempotent operation, so it is much easier to clean up
        everything and re-create them instead of parsing and figuring out what changed

        At the end, only relevant diff, if any, will be committed.
        """
        exectools.cmd_assert('git -C {} rm -rf *'.format(self.bundle_clone_path))

    def get_operator_package_yaml_info(self):
        """Get operator package name and channel from its package YAML
        This info will be used to generate bundle's Dockerfile labels and metadata/annotations.yaml
        """
        file_path = glob.glob('{}/*package.yaml'.format(self.operator_manifests_dir))[0]
        package_yaml = yaml.safe_load(io.open(file_path, encoding='utf-8'))

        self.package = package_yaml['packageName']
        self.channel = str(package_yaml['channels'][0]['name'])

    def copy_operator_manifests_to_bundle(self):
        """Copy all manifests from the operator distgit repository over to its corresponding bundle
        repository (except image-references file)
        We can be sure that the manifests contents are exactly what we expect, because our copy of
        operator repository is checked out to the specific commit used to build given operator NVR
        """
        exectools.cmd_assert('mkdir -p {}'.format(self.bundle_manifests_dir))
        exectools.cmd_assert('cp -r {} {}/'.format(
            ' '.join(self.list_of_manifest_files_to_be_copied),
            self.bundle_manifests_dir
        ))
        exectools.cmd_assert('rm -f {}/image-references'.format(self.bundle_manifests_dir))

    def replace_image_references_by_sha_on_bundle_manifests(self):
        """Iterate through all bundle manifests files, replacing any image reference tag by its
        corresponding SHA
        That is used to allow disconnected installs, where a cluster can't reach external registries
        in order to translate image tags into something "pullable"
        """
        for file in glob.glob('{}/*'.format(self.bundle_manifests_dir)):
            with io.open(file, 'r+', encoding='utf-8') as f:
                contents = self.find_and_replace_image_references_by_sha(f.read())
                f.seek(0)
                f.truncate()
                f.write(contents)

    def generate_bundle_annotations(self):
        """Create an annotations YAML file for the bundle, using info extracted from operator's
        package YAML
        """
        annotations_file = '{}/metadata/annotations.yaml'.format(self.bundle_clone_path)
        exectools.cmd_assert('mkdir -p {}'.format(os.path.dirname(annotations_file)))

        with io.open(annotations_file, 'w', encoding='utf-8') as writer:
            writer.write(yaml.dump({'annotations': self.operator_framework_tags}))

    def generate_bundle_dockerfile(self):
        """Create a Dockerfile with instructions to build the bundle container and a set of LABELs
        that allow the bundle to publish its manifests on operator's behalf.
        """
        operator_df = DockerfileParser('{}/Dockerfile'.format(self.operator_clone_path))
        bundle_df = DockerfileParser('{}/Dockerfile'.format(self.bundle_clone_path))

        bundle_df.content = 'FROM scratch\nCOPY ./manifests /manifests\nCOPY ./metadata /metadata'
        bundle_df.labels = operator_df.labels
        bundle_df.labels['com.redhat.component'] = self.bundle_brew_component
        bundle_df.labels['com.redhat.delivery.appregistry'] = False
        bundle_df.labels['name'] = 'openshift/ose-{}'.format(self.bundle_name)
        bundle_df.labels['version'] = '{}.{}'.format(
            operator_df.labels['version'],
            operator_df.labels['release']
        )
        bundle_df.labels = {
            **bundle_df.labels,
            **self.redhat_delivery_tags,
            **self.operator_framework_tags
        }
        del(bundle_df.labels['release'])

    def create_container_yaml(self):
        """Use container.yaml to disable unnecessary multiarch
        """
        filename = '{}/container.yaml'.format(self.bundle_clone_path)
        with io.open(filename, 'w', encoding='utf-8') as writer:
            writer.write('# metadata containers are not functional and do not need to be multiarch')
            writer.write('\n\n')
            writer.write(yaml.dump({
                'platforms': {'only': ['x86_64']},
                'operator_manifests': {'manifests_dir': 'manifests'},
            }))

    def commit_and_push_bundle(self, commit_msg):
        """Try to commit and push bundle distgit repository if there were any content changes.

        :param string commit_msg: Commit message
        :return bool True if new changes were committed and pushed, False otherwise
        """
        with pushd.Dir(self.bundle_clone_path):
            try:
                exectools.cmd_assert('git add .')
                exectools.cmd_assert('rhpkg{}commit -m "{}"'.format(self.rhpkg_opts, commit_msg))
                rc, out, err = exectools.cmd_gather('rhpkg{}push'.format(self.rhpkg_opts))
                return True
            except Exception:
                return False  # Bundle repository might be already up-to-date, nothing new to commit

    def trigger_bundle_container_build(self):
        """Ask brew for a container-build of operator's bundle

        :return bool True if brew task was successfully created, False otherwise
        """
        with pushd.Dir(self.bundle_clone_path):
            rc, out, err = exectools.cmd_gather(
                'rhpkg{}container-build --nowait --target {}'.format(self.rhpkg_opts, self.target)
            )

        if rc != 0:
            msg = 'Unable to create brew task: rc={} out={} err={}'.format(rc, out, err)
            self.runtime.logger.info(msg)
            return False

        self.task_url = re.search(r'Task info:\s(.+)', out).group(1)
        self.task_id = re.search(r'Created task:\s(\d+)', out).group(1)
        return True

    def watch_bundle_container_build(self):
        """Log brew task URL and eventual task states until task completion (or failure)

        :return bool True if brew task was successfully completed, False otherwise
        """
        self.runtime.logger.info('Build running: {}'.format(self.task_url))
        error = brew.watch_task(
            self.runtime.group_config.urls.brewhub,
            self.runtime.logger.info,
            self.task_id,
            threading.Event()
        )
        if error:
            self.runtime.logger.info(error)
            return False
        return True

    def find_and_replace_image_references_by_sha(self, contents):
        """Search image references (<registry>/<image>:<tag>) on given contents (usually YAML),
        replace them with corresponding (<registry>/<image>@<sha>) and collect such replacements to
        list them as "relatedImages" under "spec" section of contents (should it exist)

        :param string contents: File contents that potentially contains image references
        :return string: Same contents, with aforementioned modifications applied
        """
        found_images = {}

        def collect_replaced_image(match):
            image = '{}/{}@{}'.format(
                'registry.redhat.io',  # hardcoded until appregistry is dead
                match.group(1).replace('openshift/', 'openshift4/'),
                self.fetch_image_sha('{}:{}'.format(match.group(1), match.group(2)))
            )
            key = u'{}'.format(re.search(r'([^\/]+)\/(.+)', match.group(1)).group(2))
            found_images[key] = u'{}'.format(image)
            return image

        new_contents = re.sub(
            r'{}\/([^:]+):([^\'"\s]+)'.format(self.operator_csv_config['registry']),
            collect_replaced_image,
            contents,
            flags=re.MULTILINE
        )

        return self.append_related_images_spec(new_contents, found_images)

    def fetch_image_sha(self, image):
        """Get corresponding SHA of given image (using `oc image info`)

        OCP 4.3+ supports "manifest-lists", which is a SHA that doesn't represent an actual image,
        but a list of images per architecture instead. OCP 4.3+ is smart enough to read that list
        and pick the correct architecture.

        Unfortunately, OCP 4.2 is multi-arch and does not support manifest-lists. It is still unclear
        how we want to handle this case on this new bundle workflow. Previously it was easy to simply
        generate manifests for all arches in a single run, since all manifests were living together
        under the same branch.

        Possible solutions:
        * Build multiple sets of operator bundles, one per architecture
        Caveats: More manual work, and maybe more advisories, since I'm not sure if Errata Tool will
        let us attach multiple builds of the same brew component to a single advisory.

        * Have a multi-arch build of the bundle container
        Caveats: Not sure if appregistry/IIB will know what to do with that, publishing each arch in
        a different channel

        For now, simply assuming x86_64 (aka amd64 in golang land)

        :param string image: Image reference (format: <registry>/<image>:<tag>)
        :return string: SHA of corresponding <tag> (format: sha256:a1b2c3d4...)
        """
        registry = self.runtime.group_config.urls.brew_image_host.rstrip('/')
        ns = self.runtime.group_config.urls.brew_image_namespace
        image = '{}/{}'.format(ns, image.replace('/', '-')) if ns else image

        pull_spec = '{}/{}'.format(registry, image)
        cmd = 'oc image info --filter-by-os=linux/amd64 -o json {}'.format(pull_spec)
        try:
            out, err = exectools.cmd_assert(cmd, retries=3)
        except:
            self.runtime.logger.error(f'Unable to find image from CSV: {pull_spec}. Image may have failed to build after CSV rebase.')
            raise

        if self.runtime.group_config.operator_image_ref_mode == 'manifest-list':
            return json.loads(out)['listDigest']

        # @TODO: decide how to handle 4.2 multi-arch. hardcoding amd64 for now
        return json.loads(out)['contentDigest']

    def append_related_images_spec(self, contents, images):
        """Create a new section under contents' "spec" called "relatedImages", listing all given
        images in the following format:

        spec:
          relatedImages:
            - name: image-a
              image: registry/image-a@sha256:....
            - name: image-b
              image: registry/image-b@sha256:....

        If list of images is empty or "spec" section is not found, return contents as-is

        :param string contents: File contents that potentially contains a "spec" section
        :param list images: List of image info dictionaries (format: [{"name": "...", "image": "..."}])
        :return string: Given contents with aforementioned modifications applied
        """
        if not images:
            return contents

        related_images = [
            '    - name: {}\n      image: {}'.format(name, image)
            for name, image in images.items()
        ]
        related_images.sort()

        return re.sub(
            r'^spec:\n',
            'spec:\n  relatedImages:\n{}\n'.format('\n'.join(related_images)),
            contents,
            flags=re.MULTILINE
        )

    @property
    def brew_session(self):
        if not hasattr(self, '_brew_session'):
            self._brew_session = brew.koji.ClientSession(
                self.runtime.group_config.urls.brewhub,
                opts={'serverca': '/etc/pki/brew/legacy.crt'}
            )
        return self._brew_session

    @property
    def operator_name(self):
        return self.operator_repo_name.split('/')[-1]

    @property
    def operator_csv_config(self):
        return self.runtime.image_map[self.operator_name].config['update-csv']

    @property
    def operator_clone_path(self):
        return '{}/distgits/{}'.format(self.runtime.working_dir, self.operator_repo_name)

    @property
    def operator_manifests_dir(self):
        return '{}/{}'.format(
            self.operator_clone_path,
            self.operator_csv_config['manifests-dir'].rstrip('/')
        )

    @property
    def operator_bundle_dir(self):
        return '{}/{}'.format(
            self.operator_manifests_dir,
            self.operator_csv_config['bundle-dir'].rstrip('/')
        )

    @property
    def operator_brew_component(self):
        config = self.runtime.image_map[self.operator_name].config

        if 'distgit' in config and 'component' in config['distgit']:
            return config['distgit']['component']

        return '{}-container'.format(self.operator_name)

    @property
    def bundle_name(self):
        return '{}-bundle'.format(self.operator_name)

    @property
    def bundle_repo_name(self):
        return '{}-bundle'.format(self.operator_repo_name)

    @property
    def bundle_clone_path(self):
        return '{}-bundle'.format(self.operator_clone_path)

    @property
    def bundle_manifests_dir(self):
        return '{}/manifests'.format(self.bundle_clone_path)

    @property
    def bundle_brew_component(self):
        return self.operator_brew_component.replace('-container', '-metadata-container')

    @property
    def branch(self):
        return self.runtime.group_config.branch.format(**self.runtime.group_config.vars)

    @property
    def rhpkg_opts(self):
        opts = self.runtime.rhpkg_config
        if hasattr(self.runtime, 'user') and self.runtime.user is not None:
            opts += ' --user {} '.format(self.runtime.user)
        return opts

    @property
    def list_of_manifest_files_to_be_copied(self):
        files = glob.glob('{}/*'.format(self.operator_bundle_dir))
        if not files:
            # 4.1 channel in package YAML is "preview" or "stable", but directory name is "4.1"
            files = glob.glob('{}/{}/*'.format(
                self.operator_manifests_dir,
                '{MAJOR}.{MINOR}'.format(**self.runtime.group_config.vars)
            ))
        return files

    @property
    def redhat_delivery_tags(self):
        versions = 'v{MAJOR}.{MINOR}'.format(**self.runtime.group_config.vars)

        return {
            # 'com.redhat.delivery.backport': 'true',
            'com.redhat.delivery.operator.bundle': 'true',
            'com.redhat.openshift.versions': versions,
        }

    @property
    def operator_framework_tags(self):
        return {
            'operators.operatorframework.io.bundle.channel.default.v1': self.channel,
            'operators.operatorframework.io.bundle.channels.v1': self.channel,
            'operators.operatorframework.io.bundle.manifests.v1': 'manifests/',
            'operators.operatorframework.io.bundle.mediatype.v1': 'registry+v1',
            'operators.operatorframework.io.bundle.metadata.v1': 'metadata/',
            'operators.operatorframework.io.bundle.package.v1': self.package,
        }

    @property
    def target(self):
        return self.runtime.get_default_candidate_brew_tag() or '{}-candidate'.format(self.branch)



from logging import Logger
from typing import Dict, List, Optional, Union, Set

from koji import ClientSession

from doozerlib import brew


class EmbargoDetector:
    """ an EmbargoDetector can find builds with embargoed fixes
    """
    def __init__(self, session: ClientSession, logger: Optional[Logger] = None):
        """ creates a new EmbargoDetector
        :param session: a koji client session
        :param logger: a logger
        """
        self.koji_session = session
        self.logger = logger
        self.shipping_statuses: Dict[int, bool] = {}  # a dict for caching build shipping statues. key is build id, value is True if shipped.
        self.archive_lists: Dict[int, List[Dict]] = {}  # a dict for caching archive lists. key is build id, value is a list of archives associated with that build.

    def find_embargoed_builds(self, builds: List[Dict]) -> Set[int]:
        """ find embargoes in given list of koji builds
        :param builds: a list of koji build dicts returned by the koji api
        :return: a set of build IDs that have embargoed fixes
        """
        # first, exclude all shipped builds
        self.logger and self.logger.info("Filtering out shipped builds...")
        shipped = self.find_shipped_builds([b["id"] for b in builds])
        suspects = [b for b in builds if b["id"] not in shipped]

        # second, if a build's release field includes .p1, it is embargoed
        embargoed = {b["id"] for b in suspects if ".p1" in b["release"]}

        # finally, look at the rpms in .p0 images in case they include unshipped .p1 rpms
        suspect_build_ids = {b["id"] for b in suspects if b["id"] not in embargoed}  # non .p1 build IDs

        build_ids = suspect_build_ids - self.archive_lists.keys()  # a set of build IDs that are not in self.archive_lists cache
        if build_ids:
            build_ids = list(build_ids)
            self.logger and self.logger.info(f"Fetching image archives for {len(build_ids)} builds...")
            archive_lists = brew.list_archives_by_builds(build_ids, "image", self.koji_session)  # if a build is not an image (e.g. rpm), Brew will return an empty archive list for that build
            for index, archive_list in enumerate(archive_lists):
                self.archive_lists[build_ids[index]] = archive_list  # save to cache
        suspect_archives = [ar for suspect in suspect_build_ids for ar in self.archive_lists[suspect]]

        self.logger and self.logger.info(f'Fetching rpms in {len(suspect_archives)} images...')
        suspect_rpm_lists = brew.list_image_rpms([ar["id"] for ar in suspect_archives], self.koji_session)
        for index, rpms in enumerate(suspect_rpm_lists):
            suspected_rpms = [rpm for rpm in rpms if ".p1" in rpm["release"]]  # there should be a better way to checking the release field...
            shipped = self.find_shipped_builds([rpm["build_id"] for rpm in suspected_rpms])
            embargoed_rpms = [rpm for rpm in suspected_rpms if rpm["build_id"] not in shipped]
            if embargoed_rpms:
                image_build_id = suspect_archives[index]["build_id"]
                embargoed.add(image_build_id)
        return embargoed

    def find_shipped_builds(self, build_ids: Set[Union[int, str]]):
        """ find shipped builds in the given builds
        :param build_ids: a list of build IDs
        :return: a set of shipped build IDs
        """
        uncached = set(build_ids) - self.shipping_statuses.keys()
        if uncached:
            uncached = list(uncached)
            self.logger and self.logger.info(f'Getting tags for {len(uncached)} builds...')
            tag_lists = brew.get_builds_tags(uncached, self.koji_session)
            for index, tags in enumerate(tag_lists):
                build_id = uncached[index]
                # a shipped build should have a Brew tag ending with `-released`, like `RHBA-2020:2713-released`
                shipped = any(map(lambda tag: tag["name"].endswith("-released"), tags))
                self.shipping_statuses[build_id] = shipped  # save to cache
        result = set(filter(lambda build_id: self.shipping_statuses[build_id], build_ids))
        return result



from __future__ import absolute_import, print_function, unicode_literals
from future import standard_library
standard_library.install_aliases
from typing import Dict, Optional, List
import urllib.parse
from retry import retry
import requests
import yaml
from collections import OrderedDict
from functools import lru_cache
from dockerfile_parse import DockerfileParser
import pathlib
import json
import traceback
import glob

from .pushd import Dir
from .distgit import ImageDistGitRepo, RPMDistGitRepo, DistGitRepo
from . import exectools
from . import logutil

from .model import Model, Missing

#
# These are used as labels to index selection of a subclass.
#
DISTGIT_TYPES = {
    'image': ImageDistGitRepo,
    'rpm': RPMDistGitRepo
}

CONFIG_MODES = [
    'enabled',  # business as usual
    'disabled',  # manually disabled from automatically building
    'wip',  # Work in Progress, do not build
]

CONFIG_MODE_DEFAULT = CONFIG_MODES[0]


class Metadata(object):

    def __init__(self, meta_type, runtime, data_obj: Dict, commitish: Optional[str] = None):
        """
        :param: meta_type - a string. Index to the sub-class <'rpm'|'image'>.
        :param: runtime - a Runtime object.
        :param: name - a filename to load as metadata
        :param: commitish: If not None, build from the specified upstream commit-ish instead of the branch tip.
        """
        self.meta_type = meta_type
        self.runtime = runtime
        self.data_obj = data_obj
        self.config_filename = data_obj.filename
        self.full_config_path = data_obj.path
        self.commitish = commitish

        # URL and branch of public upstream source are set later by Runtime.resolve_source()
        self.public_upstream_url = None
        self.public_upstream_branch = None

        # Some config filenames have suffixes to avoid name collisions; strip off the suffix to find the real
        # distgit repo name (which must be combined with the distgit namespace).
        # e.g. openshift-enterprise-mediawiki.apb.yml
        #      distgit_key=openshift-enterprise-mediawiki.apb
        #      name (repo name)=openshift-enterprise-mediawiki

        self.distgit_key = data_obj.key
        self.name = self.distgit_key.split('.')[0]   # Split off any '.apb' style differentiator (if present)

        self.runtime.logger.debug("Loading metadata from {}".format(self.full_config_path))

        self.config = Model(data_obj.data)

        self.mode = self.config.get('mode', CONFIG_MODE_DEFAULT).lower()
        if self.mode not in CONFIG_MODES:
            raise ValueError('Invalid mode for {}'.format(self.config_filename))

        self.enabled = (self.mode == CONFIG_MODE_DEFAULT)

        # Basic config validation. All images currently required to have a name in the metadata.
        # This is required because from.member uses these data to populate FROM in images.
        # It would be possible to query this data from the distgit Dockerflie label, but
        # not implementing this until we actually need it.
        assert (self.config.name is not Missing)

        # Choose default namespace for config data
        if meta_type == "image":
            self.namespace = "containers"
        else:
            self.namespace = "rpms"

        # Allow config data to override
        if self.config.distgit.namespace is not Missing:
            self.namespace = self.config.distgit.namespace

        self.qualified_name = "%s/%s" % (self.namespace, self.name)
        self.qualified_key = "%s/%s" % (self.namespace, self.distgit_key)

        # Includes information to identify the metadata being used with each log message
        self.logger = logutil.EntityLoggingAdapter(logger=self.runtime.logger, extra={'entity': self.qualified_key})

        self._distgit_repo = None

    def save(self):
        self.data_obj.data = self.config.primitive()
        self.data_obj.save()

    def distgit_remote_url(self):
        pkgs_host = self.runtime.group_config.urls.get('pkgs_host', 'pkgs.devel.redhat.com')
        # rhpkg uses a remote named like this to pull content from distgit
        if self.runtime.user:
            return f'ssh://{self.runtime.user}@{pkgs_host}/{self.qualified_name}'
        return f'ssh://{pkgs_host}/{self.qualified_name}'

    def distgit_repo(self, autoclone=True) -> DistGitRepo:
        if self._distgit_repo is None:
            self._distgit_repo = DISTGIT_TYPES[self.meta_type](self, autoclone=autoclone)
        return self._distgit_repo

    def branch(self):
        if self.config.distgit.branch is not Missing:
            return self.config.distgit.branch
        return self.runtime.branch

    def candidate_brew_tag(self):
        return '{}-candidate'.format(self.branch())

    def get_arches(self):
        """
        :return: Returns the list of architecture this image/rpm should build for. This is an intersection
        of config specific arches & globally enabled arches in group.yml
        """
        if self.config.arches:
            ca = self.config.arches
            intersection = list(set(self.runtime.get_global_arches()) & set(ca))
            if len(intersection) != len(ca):
                self.logger.info(f'Arches are being pruned by group.yml. Using computed {intersection} vs config list {ca}')
            if not intersection:
                raise ValueError(f'No arches remained enabled in {self.qualified_key}')
            return intersection
        else:
            return list(self.runtime.get_global_arches())

    def cgit_url(self, filename):
        rev = self.branch()
        ret = "/".join((self.runtime.group_config.urls.cgit, self.qualified_name, "plain", filename))
        if rev is not None:
            ret = "{}?h={}".format(ret, rev)
        return ret

    def fetch_cgit_file(self, filename):
        url = self.cgit_url(filename)
        req = exectools.retry(
            3, lambda: urllib.request.urlopen(url),
            check_f=lambda req: req.code == 200)
        return req.read()

    def get_latest_build(self, default=-1):
        """
        :param default: A value to return if no latest is found (if not specified, an exception will be thrown)
        :return: Returns the most recent build object from koji.
                 Example https://gist.github.com/jupierce/e6bfd98a3777ae5d56e0f7e92e5db0c9
        """
        component_name = self.get_component_name()
        with self.runtime.pooled_koji_client_session() as koji_api:
            builds = koji_api.getLatestBuilds(self.candidate_brew_tag(), package=component_name)
            if not builds:
                if default != -1:
                    self.logger.warning("No builds detected for using tag: %s" % (self.candidate_brew_tag()))
                    return default
                raise IOError("No builds detected for %s using tag: %s" % (self.qualified_name, self.candidate_brew_tag()))
            return builds[0]

    def get_latest_build_info(self, default=-1):
        """
        Queries brew to determine the most recently built release of the component
        associated with this image. This method does not rely on the "release"
        label needing to be present in the Dockerfile.
        :param default: A value to return if no latest is found (if not specified, an exception will be thrown)
        :return: A tuple: (component name, version, release); e.g. ("registry-console-docker", "v3.6.173.0.75", "1")
        """
        build = self.get_latest_build(default=default)
        if default != -1 and build == default:
            return default
        return build['name'], build['version'], build['release']

    def get_component_name(self, default=-1):
        """
        :param default: If the component name cannot be determine,
        :return: Returns the component name of the image. This is the name in the nvr
        that brew assigns to this image's build.
        """
        raise IOError('Subclass must implement')

    def needs_rebuild(self):
        """
        Check whether the commit that we recorded in the distgit content (according to cgit)
        matches the commit of the source (according to git ls-remote) and has been built
        (according to brew). Nothing is cloned and no existing clones are consulted.
        Returns: (<bool>, message). If True, message describing the details is returned. If False,
                None is returned.
        """
        component_name = self.get_component_name(default='')
        if not component_name:
            # This can happen for RPMs if they have never been rebased into
            # distgit.
            return True, 'Could not find component name; assuming never built'

        latest_build = self.get_latest_build(default='')
        if not latest_build:
            # Truly never built.
            return True, f'Component {component_name} has never been built'

        latest_build_creation_event_id = latest_build['creation_event_id']
        with self.runtime.pooled_koji_client_session() as koji_api:
            # getEvent returns something like {'id': 31825801, 'ts': 1591039601.2667}
            latest_build_creation_ts = int(koji_api.getEvent(latest_build_creation_event_id)['ts'])

        dgr = self.distgit_repo()
        with Dir(dgr.distgit_dir):
            ts, _ = exectools.cmd_assert('git show -s --format=%ct HEAD')
            distgit_head_commit_millis = int(ts)

        one_hour = (1 * 60 * 60 * 1000)  # in milliseconds

        if not dgr.has_source():
            if distgit_head_commit_millis > latest_build_creation_ts:
                # Two options:
                # 1. A user has made a commit to this dist-git only branch and there has been no build attempt
                # 2. We've already tried a build and the build failed.
                # To balance these two options, if the diff > 1 hour, request a build.
                if (distgit_head_commit_millis - latest_build_creation_ts) > one_hour:
                    return True, 'Distgit only repo commit is at least one hour older than most recent build'
            return False, 'Distgit only repo commit is older than most recent build'

        # We have source.
        with Dir(dgr.source_path()):
            upstream_commit_hash, _ = exectools.cmd_assert('git rev-parse HEAD', strip=True)

        dgr_path = pathlib.Path(dgr.distgit_dir)
        if self.namespace == 'containers' or self.namespace == 'apbs':
            dockerfile_path = dgr_path.joinpath('Dockerfile')
            if not dockerfile_path.is_file():
                return True, 'Distgit dockerfile not found -- appears that no rebase has ever been performed'
            dfp = DockerfileParser(str(dockerfile_path))
            last_disgit_rebase_hash = dfp.envs.get('SOURCE_GIT_COMMIT', None)
            if last_disgit_rebase_hash != upstream_commit_hash:
                return True, f'Distgit contains SOURCE_GIT_COMMIT hash {last_disgit_rebase_hash} different from upstream HEAD {upstream_commit_hash}'
        elif self.namespace == 'rpms':
            specs = list(dgr_path.glob('*.spec'))
            if not specs:
                return True, 'Distgit .spec file not found -- appears that no rebase has ever been performed'
            with specs[0].open(mode='r', encoding='utf-8') as spec_handle:
                spec_content = spec_handle.read()
                if upstream_commit_hash not in spec_content:
                    return True, f'Distgit spec file does not contain upstream hash {upstream_commit_hash}'
        else:
            raise IOError(f'Unknown namespace type: {self.namespace}')

        if distgit_head_commit_millis > latest_build_creation_ts:
            # Distgit is ahead of the latest build.
            # We've likely made an attempt to rebase and the subsequent build failed.
            # Try again if we are at least 6 hours out from the build to avoid
            # pestering image owners will repeated build failures.
            if distgit_head_commit_millis - latest_build_creation_ts > (6 * one_hour):
                return True, 'It has been 6 hours since last failed build attempt'
            return False, f'Distgit commit ts {distgit_head_commit_millis} ahead of last successful build ts {latest_build_creation_ts}, but holding off for at least 6 hours before rebuild'
        else:
            # The latest build is newer than the latest distgit commit. No change required.
            return False, 'Latest build is newer than latest upstream/distgit commit -- no build required'

    def get_maintainer_info(self):
        """
        :return: Returns a dict of identifying maintainer information. Dict might be empty if no maintainer information is available.
            fields are generally [ component: '...', subcomponent: '...', and product: '...' ] if available. These
            are coordinates for product security to figure out where to file bugs when an image or RPM has an issue.
        """

        # We are trying to discover some team information that indicates which BZ or Jira board bugs for this
        # component should be filed against. This information can be stored in the doozer metadata OR
        # in upstream source. Metadata overrides, as usual.

        source_dir = self.runtime.resolve_source(self)

        # Maintainer info can be defined in metadata, so try there first.
        maintainer = self.config.maintainer or dict()

        # This tuple will also define key ordering in the returned OrderedDict
        known_fields = ('product', 'component', 'subcomponent')

        # Fill in any missing attributes from upstream source
        if source_dir:
            with Dir(source_dir):
                # Not every repo has a master branch, they may have a different default; detect it.
                remote_info, _ = exectools.cmd_assert('git remote show origin')
                head_branch_lines = [i for i in remote_info.splitlines() if i.strip().startswith('HEAD branch:')]  # e.g. [ "  HEAD branch: master" ]
                if not head_branch_lines:
                    raise IOError('Error trying to detect remote default branch')
                default_branch = head_branch_lines[0].strip().split()[-1]  # [ "  HEAD branch: master" ] => "master"

                _, owners_yaml, _ = exectools.cmd_gather(f'git --no-pager show origin/{default_branch}:OWNERS')
                if owners_yaml.strip():
                    owners = yaml.safe_load(owners_yaml)
                    for field in known_fields:
                        if field not in maintainer and field in owners:
                            maintainer[field] = owners[field]

        if 'product' not in maintainer:
            maintainer['product'] = 'OpenShift Container Platform'  # Safe bet - we are ART.

        # Just so we return things in a defined order (avoiding unnecessary changes in git commits)
        sorted_maintainer = OrderedDict()
        for k in known_fields:
            if k in maintainer:
                sorted_maintainer[k] = maintainer[k]

        # Add anything remaining in alpha order
        for k in sorted(maintainer.keys()):
            if k not in sorted_maintainer:
                sorted_maintainer[k] = maintainer[k]

        return sorted_maintainer

    def extract_kube_env_vars(self) -> Dict[str, str]:
        """
        Analyzes the source_base_dir for either Godeps or go.mod and looks for information about
        which version of Kubernetes is being utilized by the repository. Side effect is cloning distgit
        and upstream source if it has not already been done.
        :return: A Dict of environment variables that should be added to the Dockerfile / rpm spec.
                Variables like KUBE_GIT_VERSION, KUBE_GIT_COMMIT, KUBE_GIT_MINOR, ...
                May be empty if there is no kube information in the source dir.
        """
        envs = dict()

        upstream_source_path: pathlib.Path = self.distgit_repo(autoclone=True).source_repo_path()
        if not upstream_source_path:
            # distgit only. Return empty.
            return envs

        kube_version_fields: List[str] = None  # Populate ['x', 'y', 'z'] this from godeps or gomod
        kube_commit_hash: str = None  # Populate with kube repo hash like '2f054b7646dc9e98f6dea458d2fb65e1d2c1f731'
        with Dir(upstream_source_path):
            out, _ = exectools.cmd_assert(["git", "rev-parse", "HEAD"])
            source_full_sha = out.strip()

            # First determine if this source repository is using Godeps. Godeps is ultimately
            # being replaced by gomod, but older versions of OpenShift continue to use it.
            godeps_file = pathlib.Path(upstream_source_path, 'Godeps', 'Godeps.json')
            if godeps_file.is_file():
                try:
                    with godeps_file.open('r', encoding='utf-8') as f:
                        godeps = json.load(f)
                        # Reproduce https://github.com/openshift/origin/blob/6f457bc317f8ca8e514270714db6597ec1cb516c/hack/lib/build/version.sh#L82
                        # Example of what we are after: https://github.com/openshift/origin/blob/6f457bc317f8ca8e514270714db6597ec1cb516c/Godeps/Godeps.json#L10-L15
                        for dep in godeps.get('Deps', []):
                            if dep.get('ImportPath', '') == 'k8s.io/kubernetes/pkg/api':
                                kube_commit_hash = dep.get('Rev', '')
                                raw_kube_version = dep.get('Comment', '')  # e.g. v1.14.6-152-g117ba1f
                                # drop release information.
                                base_kube_version = raw_kube_version.split('-')[0]  # v1.17.1-152-g117ba1f => v1.17.1
                                kube_version_fields = base_kube_version.lstrip('v').split('.')  # v1.17.1 => [ '1', '17', '1']
                except:
                    self.logger.error(f'Error parsing godeps {str(godeps_file)}')
                    traceback.print_exc()

            go_sum_file = pathlib.Path(upstream_source_path, 'go.sum')
            if go_sum_file.is_file():
                try:
                    # we are looking for a line like: https://github.com/openshift/kubernetes/blob/5241b27b8acd73cdc99a0cac281645189189f1d8/go.sum#L602
                    # e.g. "k8s.io/kubernetes v1.19.0-rc.2/go.mod h1:zomfQQTZYrQjnakeJi8fHqMNyrDTT6F/MuLaeBHI9Xk="
                    with go_sum_file.open('r', encoding='utf-8') as f:
                        for line in f.readlines():
                            if line.startswith('k8s.io/kubernetes '):
                                entry_split = line.split()  # => ['k8s.io/kubernetes', 'v1.19.0-rc.2/go.mod', 'h1:zomfQQTZYrQjnakeJi8fHqMNyrDTT6F/MuLaeBHI9Xk=']
                                base_kube_version = entry_split[1].split('/')[0].strip()  # 'v1.19.0-rc.2/go.mod' => 'v1.19.0-rc.2'
                                kube_version_fields = base_kube_version.lstrip('v').split('.')  # 'v1.19.0-rc.2' => [ '1', '19', '0-rc.2']
                                # upstream kubernetes creates a tag for each version. Go find its sha.
                                rc, out, err = exectools.cmd_gather('git ls-remote https://github.com/kubernetes/kubernetes {base_kube_version}')
                                out = out.strip()
                                if out:
                                    # Expecting something like 'a26dc584ac121d68a8684741bce0bcba4e2f2957	refs/tags/v1.19.0-rc.2'
                                    kube_commit_hash = out.split()[0]
                                else:
                                    # That's strange, but let's not kill the build for it.  Poke in our repo's hash.
                                    kube_commit_hash = source_full_sha
                                break
                except:
                    self.logger.error(f'Error parsing go.sum {str(go_sum_file)}')
                    traceback.print_exc()

            if kube_version_fields:
                # For historical consistency with tito's flow, we add +OS_GIT_COMMIT[:7] to the kube version
                envs['KUBE_GIT_VERSION'] = f"v{'.'.join(kube_version_fields)}+{source_full_sha[:7]}"
                envs['KUBE_GIT_MAJOR'] = '0' if len(kube_version_fields) < 1 else kube_version_fields[0]
                godep_kube_minor = '0' if len(kube_version_fields) < 2 else kube_version_fields[1]
                envs['KUBE_GIT_MINOR'] = f'{godep_kube_minor}+'  # For historical reasons, append a '+' since OCP patches its vendored kube.
                envs['KUBE_GIT_COMMIT'] = kube_commit_hash
                envs['KUBE_GIT_TREE_STATE'] = 'clean'
            elif self.name in ('openshift-enterprise-hyperkube', 'openshift', 'atomic-openshift'):
                self.logger.critical(f'Unable to acquire KUBE vars for {self.name}. This must be fixed or platform addons can break: https://bugzilla.redhat.com/show_bug.cgi?id=1861097')
                raise IOError(f'Unable to determine KUBE vars for {self.name}')

            return envs



import click
import yaml

from doozerlib import brew, rhcos, exectools
from doozerlib.cli import cli, pass_runtime
from doozerlib.cli import release_gen_payload as rgp
from doozerlib.exceptions import DoozerFatalError


@cli.command("config:scan-sources", short_help="Determine if any rpms / images need to be rebuilt.")
@click.option("--ci-kubeconfig", metavar='KC_PATH', required=False,
              help="File containing kubeconfig for looking at release-controller imagestreams")
@click.option("--yaml", "as_yaml", default=False, is_flag=True, help='Print results in a yaml block')
@pass_runtime
def config_scan_source_changes(runtime, ci_kubeconfig, as_yaml):
    """
    Determine if any rpms / images need to be rebuilt.

    \b
    The method will report RPMs in this group if:
    - Their source git hash no longer matches their upstream source.
    - The buildroot used by the previous RPM build has changed.

    \b
    It will report images if the latest build:
    - Contains an RPM that is about to be rebuilt based on the RPM check above.
    - If the source git hash no longer matches the upstream source.
    - Contains any RPM (from anywhere in Red Hat) which has likely changed since the image was built.
        - This indirectly detects non-member parent image changes.
    - Was built with a buildroot that has now changed (probably not useful for images, but was cheap to add).
    - Used a builder image (from anywhere in Red Hat) that has changed.
    - Used a builder image from this group that is about to change.
    - If the associated member is a descendant of any image that needs change.

    \b
    It will report machine-os-content updates available per imagestream.
    """
    runtime.initialize(mode='both', clone_distgits=True)

    all_rpm_metas = set(runtime.rpm_metas())
    all_image_metas = set(runtime.image_metas())

    changing_rpm_metas = set()
    changing_image_metas = set()
    changing_rpm_packages = set()
    assessment_reason = dict()  # maps metadata qualified_key => message describing change

    def add_assessment_reason(meta, changing, reason):
        # qualify by whether this is a True or False for change so that we can store both in the map.
        key = f'{meta.qualified_key}+{changing}'
        # If the key is already there, don't replace the message as it is likely more interesting
        # than subsequent reasons (e.g. changing because of ancestry)
        if key not in assessment_reason:
            assessment_reason[key] = reason

    with runtime.shared_koji_client_session() as koji_api:
        runtime.logger.info(f'Running scan with true latest koji/brew event: {koji_api.getLastEvent()}')

        # Different branches have different build tags & inheritances; cache the results to
        # limit brew queries.
        build_root_tag_id_cache = dict()

        def get_build_root_inherited_tags(config_meta):
            # Create a {map of tags => tag names} which contribute RPMs to a given build root
            build_tag = config_meta.branch() + '-build'
            if build_tag in build_root_tag_id_cache:
                return build_root_tag_id_cache[build_tag]
            build_tag_ids = [n['parent_id'] for n in koji_api.getFullInheritance(build_tag)]
            final_build_tag_ids = {}
            for bid in build_tag_ids:
                tag_name = koji_api.getTag(bid)['name']
                if tag_name.endswith(('-candidate', '-pending')):
                    # We don't want our candidate tag to be taken in consideration. It will always
                    # be changing as after any build.
                    continue

                # There are many tags in the buildroots that are unlikely to affect our builds.
                # (e.g. rhel-7.8-z). Filter these down to those which reasonably affect us.
                if any(('devtools' in tag_name,  # e.g. devtools-2019.4-rhel-7
                        '-build' in tag_name,
                        '-override' in tag_name)):
                    final_build_tag_ids[bid] = tag_name

            runtime.logger.info(f'For {config_meta.distgit_key}, found buildroot inheritance: {build_tag_ids}; Filtered down to {final_build_tag_ids}')
            build_root_tag_id_cache[build_tag] = final_build_tag_ids
            return final_build_tag_ids

        # First, scan for any upstream source code changes. If found, these are guaranteed rebuilds.
        for meta, change_info in runtime.scan_distgit_sources():
            needs_rebuild, reason = change_info
            dgk = meta.distgit_key
            if not (meta.enabled or meta.mode == "disabled" and runtime.load_disabled):
                # An enabled image's dependents are always loaded. Ignore disabled configs unless explicitly indicated
                continue

            if meta.meta_type == 'rpm':
                package_name = meta.get_package_name(default=None)
                if needs_rebuild is False and package_name:  # If we are currently matching, check buildroots to see if it unmatches us
                    for latest_rpm_build in koji_api.getLatestRPMS(tag=meta.branch() + '-candidate', package=package_name)[1]:
                        # Detect if our buildroot changed since the last build of this rpm
                        rpm_build_root_tag_ids = get_build_root_inherited_tags(meta).keys()
                        build_root_changes = brew.tags_changed_since_build(runtime, koji_api, latest_rpm_build, rpm_build_root_tag_ids)
                        if build_root_changes:
                            changing_tag_names = [brc['tag_name'] for brc in build_root_changes]
                            reason = f'Latest rpm was built before buildroot changes: {changing_tag_names}'
                            runtime.logger.info(f'{dgk} ({latest_rpm_build}) will be rebuilt because it has not been built since a buildroot change: {build_root_changes}')
                            needs_rebuild = True

                if reason:
                    add_assessment_reason(meta, needs_rebuild, reason=reason)

                if needs_rebuild:
                    changing_rpm_metas.add(meta)
                    if package_name:
                        changing_rpm_packages.add(package_name)
                    else:
                        runtime.logger.warning(f"Appears that {dgk} has never been built before; can't determine package name")
            elif meta.meta_type == 'image':
                if reason:
                    add_assessment_reason(meta, needs_rebuild, reason=reason)
                if needs_rebuild:
                    changing_image_metas.add(meta)
            else:
                raise IOError(f'Unsupported meta type: {meta.meta_type}')

        def add_image_meta_change(meta, msg):
            nonlocal changing_image_metas
            changing_image_metas.add(meta)
            add_assessment_reason(meta, True, msg)
            for descendant_meta in meta.get_descendants():
                changing_image_metas.add(descendant_meta)
                add_assessment_reason(descendant_meta, True, f'Ancestor {meta.distgit_key} is changing')

        # To limit the size of the queries we are going to make, find the oldest image
        eldest_image_event_ts = koji_api.getLastEvent()['ts']
        for image_meta in runtime.image_metas():
            info = image_meta.get_latest_build(default=None)
            if info is not None:
                create_event_ts = koji_api.getEvent(info['creation_event_id'])['ts']
                if create_event_ts < eldest_image_event_ts:
                    eldest_image_event_ts = create_event_ts

        for image_meta in runtime.image_metas():
            image_change, msg = image_meta.does_image_need_change(eldest_image_event_ts, changing_rpm_packages, get_build_root_inherited_tags(image_meta).keys())
            if image_change:
                add_image_meta_change(image_meta, msg)

        # does_image_name_change() checks whether its non-member builder images have changed
        # but cannot determine whether member builder images have changed until anticipated
        # changes have been calculated. The following loop does this. It uses while True,
        # because technically, we could keep finding changes in intermediate builder images
        # and have to ensure we pull in images that rely on them in the next iteration.
        # fyi, changes in direct parent images should be detected by RPM changes, which
        # does does_image_name_change will detect.
        while True:
            changing_image_dgks = [meta.distgit_key for meta in changing_image_metas]
            for image_meta in all_image_metas:
                dgk = image_meta.distgit_key
                if dgk in changing_image_dgks:  # Already in? Don't look any further
                    continue

                for builder in image_meta.config['from'].builder:
                    if builder.member and builder.member in changing_image_dgks:
                        runtime.logger.info(f'{dgk} will be rebuilt due to change in builder member ')
                        add_image_meta_change(image_meta, f'Builder group member has changed: {builder.member}')

            if len(changing_image_metas) == len(changing_image_dgks):
                # The for loop didn't find anything new, we can break
                break

        # We have our information. Now build the output report..
        image_results = []
        changing_image_dgks = [meta.distgit_key for meta in changing_image_metas]
        for image_meta in all_image_metas:
            dgk = image_meta.distgit_key
            is_changing = dgk in changing_image_dgks
            image_results.append({
                'name': dgk,
                'changed': is_changing,
                'reason': assessment_reason.get(f'{image_meta.qualified_key}+{is_changing}', 'No change detected'),
            })

        rpm_results = []
        changing_rpm_dgks = [meta.distgit_key for meta in changing_rpm_metas]
        for rpm_meta in all_rpm_metas:
            dgk = rpm_meta.distgit_key
            is_changing = dgk in changing_rpm_dgks
            rpm_results.append({
                'name': dgk,
                'changed': is_changing,
                'reason': assessment_reason.get(f'{rpm_meta.qualified_key}+{is_changing}', 'No change detected'),
            })

        results = dict(
            rpms=rpm_results,
            images=image_results
        )

        if ci_kubeconfig:  # we can determine m-os-c needs updating if we can look at imagestreams
            results['rhcos'] = _detect_rhcos_status(runtime, ci_kubeconfig)

        if as_yaml:
            click.echo('---')
            click.echo(yaml.safe_dump(results, indent=4))
            return

        for kind, items in results.items():
            if not items:
                continue
            click.echo(kind.upper() + ":")
            for item in items:
                click.echo('  {} is {} (reason: {})'.format(item['name'],
                                                            'changed' if item['changed'] else 'the same',
                                                            item['reason']))


def _detect_rhcos_status(runtime, kubeconfig) -> list:
    """
    gather the existing machine-os-content tags and compare them to latest rhcos builds
    @return a list of status entries like:
        {
            'name': "4.2-x86_64-priv",
            'changed': False,
            'reason': "could not find an RHCOS build to sync",
        }
    """
    statuses = []

    version = runtime.get_minor_version()
    for arch in runtime.arches:
        for private in (False, True):
            name = f"{version}-{arch}{'-priv' if private else ''}"
            try:
                tagged_mosc_id = _tagged_mosc_id(kubeconfig, version, arch, private)
                latest_rhcos_id = _latest_rhcos_build_id(version, arch, private)
                status = dict(name=name)
                if not latest_rhcos_id:
                    status['changed'] = False
                    status['reason'] = "could not find an RHCOS build to sync"
                elif tagged_mosc_id == latest_rhcos_id:
                    status['changed'] = False
                    status['reason'] = f"latest RHCOS build is still {latest_rhcos_id} -- no change from istag"
                else:
                    status['changed'] = True
                    status['reason'] = f"latest RHCOS build is {latest_rhcos_id} which differs from istag {tagged_mosc_id}"
                statuses.append(status)
            except Exception as ex:
                # don't let flakiness in rhcos lookups prevent us from scanning regular builds;
                # if anything else changed it will sync anyway.
                runtime.logger.warning(f"could not determine RHCOS status for {name}: {ex}")

    return statuses


def _tagged_mosc_id(kubeconfig, version, arch, private) -> str:
    """determine what the most recently tagged machine-os-content is in given imagestream"""
    base_name = rgp.default_is_base_name(version)
    base_namespace = rgp.default_is_base_namespace()
    name, namespace = rgp.is_name_and_space(base_name, base_namespace, arch, private)
    stdout, _ = exectools.cmd_assert(
        f"oc --kubeconfig '{kubeconfig}' --namespace '{namespace}' get istag '{name}:machine-os-content'"
        " --template '{{.image.dockerImageMetadata.Config.Labels.version}}'",
        retries=3,
        pollrate=5,
        strip=True,
    )
    return stdout if stdout else None


def _latest_rhcos_build_id(version, arch, private) -> str:
    """wrapper to return None if anything goes wrong, which will be taken as no change"""
    try:
        return rhcos.latest_rhcos_build_id(version, arch, private)
    except Exception:
        return None


cli.add_command(config_scan_source_changes)



"""
This file contains the definition of a class that can be used like shell
pushd and popd.

The Dir object is a context_manager that can be used with the Python 'with'
clause.  The context manager idiom allows the user to execute some commands
in a working directory other than the CWD and return without needing to
explicitly handle it.

Example:

  # os.getcwd() returns /tmp/somewhere
  with Dir("/tmp/somewhere/else"):
      # os.getcwd() returns /tmp/somewhere/else
      ....

  # os.getcwd() returns /tmp/somewhere
"""
from __future__ import absolute_import, print_function, unicode_literals

import os
import threading
import pathlib


class Dir(object):
    """
    Context manager to handle directory changes safely.

    On `__enter__`, `chdir`s to the given directory and on `__exit__`, `chdir`s
    back to restore the previous `cwd`.

    The current directory is also kept on thread-local storage and can be
    accessed (e.g. by multi-threaded programs that cannot rely on `chdir`) via
    the `getcwd` static method.

    The `assert_exec` and `gather_exec` member functions use the directory in
    effect automatically.
    """
    _tl = threading.local()

    def __init__(self, newdir):
        self.dir = str(newdir)
        self.previous_dir = None

    def __enter__(self):
        self.previous_dir = self.getcwd()
        os.chdir(self.dir)
        self._tl.cwd = self.dir
        return self.dir

    def __exit__(self, *args):
        os.chdir(self.previous_dir)
        self._tl.cwd = self.previous_dir

    @classmethod
    def getcwd(cls):
        """
        Provide a context dependent current working directory. This method
        will return the directory currently holding the lock.
        """
        if not hasattr(cls._tl, "cwd"):
            cls._tl.cwd = os.getcwd()
        return cls._tl.cwd

    @classmethod
    def getpath(cls):
        """
        Provide a context dependent current working directory. This method
        will return a pathlib.Path for the directory currently holding the lock.
        """
        return pathlib.Path(Dir.getcwd())



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from setuptools import setup, find_packages
import sys
if sys.version_info < (3, 6):
    sys.exit('Sorry, Python < 3.6 is not supported.')

with open('./requirements.txt') as f:
    INSTALL_REQUIRES = f.read().splitlines()


def _get_version():
    from os.path import abspath, dirname, join
    filename = join(dirname(abspath(__file__)), 'doozerlib', 'VERSION')
    return open(filename).read().strip()


setup(
    name="rh-doozer",
    author="AOS ART Team",
    author_email="aos-team-art@redhat.com",
    version=_get_version(),
    description="CLI tool for managing and automating Red Hat software releases",
    long_description=open('README.md').read(),
    url="https://github.com/openshift/doozer",
    license="Apache License, Version 2.0",
    packages=find_packages(exclude=["tests", "tests.*", "functional_tests", "functional_tests.*"]),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'doozer = doozerlib.cli.__main__:main'
        ]
    },
    install_requires=INSTALL_REQUIRES,
    test_suite='tests',
    dependency_links=[],
    python_requires='>=3.6',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Environment :: Console",
        "Operating System :: POSIX",
        "License :: OSI Approved :: Apache Software License",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "Natural Language :: English",
    ]
)



import io
import sys
from logging import Logger
from typing import Dict, List, Optional, Set

import click
import koji
import yaml
import json

from doozerlib import brew, state, exectools, embargo_detector, rhcos
from doozerlib.cli import cli, pass_runtime
from doozerlib.exceptions import DoozerFatalError
from doozerlib.image import ImageMetadata
from doozerlib.util import red_print, yellow_print


@cli.command("release:gen-payload", short_help="Generate input files for release mirroring")
@click.option("--is-name", metavar='NAME', required=False,
              help="ImageStream .metadata.name value. For example '4.2-art-latest'")
@click.option("--is-namespace", metavar='NAMESPACE', required=False,
              help="ImageStream .metadata.namespace value. For example 'ocp'")
@click.option("--organization", metavar='ORGANIZATION', required=False, default='openshift-release-dev',
              help="Quay ORGANIZATION to mirror into.\ndefault=openshift-release-dev")
@click.option("--repository", metavar='REPO', required=False, default='ocp-v4.0-art-dev',
              help="Quay REPOSITORY in ORGANIZATION to mirror into.\ndefault=ocp-v4.0-art-dev")
@click.option("--event-id", metavar='NUM', required=False, type=int,
              help="A Brew event ID. If specified, the latest images as of the given Brew event will be chosen for mirroring intead of now.")
@pass_runtime
def release_gen_payload(runtime, is_name, is_namespace, organization, repository, event_id):
    """Generates two sets of input files for `oc` commands to mirror
content and update image streams. Files are generated for each arch
defined in ocp-build-data for a version, as well as a final file for
manifest-lists.

One set of files are SRC=DEST mirroring definitions for 'oc image
mirror'. They define what source images we will sync to which
destination repos, and what the mirrored images will be labeled as.

The other set of files are YAML image stream tags for 'oc
apply'. Those are applied to an openshift cluster to define "release
streams". When they are applied the release controller notices the
update and begins generating a new payload with the images tagged in
the image stream.

For automation purposes this command generates a mirroring yaml files
after the arch-specific files have been generated. The yaml files
include names of generated content.

You may provide the namespace and base name for the image streams, or defaults
will be used. The generated files will append the -arch and -priv suffixes to
the given name and namespace as needed.

The ORGANIZATION and REPOSITORY options are combined into
ORGANIZATION/REPOSITORY when preparing for mirroring.

Generate files for mirroring from registry-proxy (OSBS storage) to our
quay registry:

\b
    $ doozer --group=openshift-4.2 release:gen-payload \\
        --is-name=4.2-art-latest

Note that if you use -i to include specific images, you should also include
openshift-enterprise-cli to satisfy any need for the 'cli' tag. The cli image
is used automatically as a stand-in for images when an arch does not build
that particular tag.
    """
    runtime.initialize(clone_distgits=False, config_excludes='non_release')
    orgrepo = "{}/{}".format(organization, repository)
    cmd = runtime.command
    runtime.state[cmd] = dict(state.TEMPLATE_IMAGE)
    lstate = runtime.state[cmd]  # get local convenience copy

    if not is_name:
        is_name = default_is_base_name(runtime.get_minor_version())
    if not is_namespace:
        is_namespace = default_is_base_namespace()

    images = [i for i in runtime.image_metas()]
    lstate['total'] = len(images)

    no_build_items = []
    invalid_name_items = []

    payload_images = []
    for image in images:
        # Per clayton:
        """Tim Bielawa: note to self: is only for `ose-` prefixed images
        Clayton Coleman: Yes, Get with the naming system or get out of town
        """
        if image.is_payload:
            if not image.image_name_short.startswith("ose-"):
                invalid_name_items.append(image.image_name_short)
                red_print("NOT adding to IS (does not meet name/version conventions): {}".format(image.image_name_short))
                continue
            else:
                payload_images.append(image)

    runtime.logger.info("Fetching latest image builds from Brew...")
    tag_component_tuples = [(image.candidate_brew_tag(), image.get_component_name()) for image in payload_images]
    brew_session = runtime.build_retrying_koji_client()
    latest_builds = brew.get_latest_builds(tag_component_tuples, "image", event_id, brew_session)
    latest_builds = [builds[0] if builds else None for builds in latest_builds]  # flatten the data structure

    runtime.logger.info("Fetching image archives...")
    build_ids = [b["id"] if b else 0 for b in latest_builds]
    archives_list = brew.list_archives_by_builds(build_ids, "image", brew_session)

    mismatched_siblings = find_mismatched_siblings(payload_images, latest_builds, archives_list, runtime.logger, lstate)

    embargoed_build_ids = set()  # a set of private image build ids
    if runtime.group_config.public_upstreams:
        # looking for embargoed image builds
        detector = embargo_detector.EmbargoDetector(brew_session, runtime.logger)
        for index, archive_list in enumerate(archives_list):
            if build_ids[index]:
                detector.archive_lists[build_ids[index]] = archive_list  # store to EmbargoDetector cache to limit Brew queries
        suspects = [b for b in latest_builds if b]
        embargoed_build_ids = detector.find_embargoed_builds(suspects)

    runtime.logger.info("Creating mirroring lists...")

    # These will map[arch] -> map[image_name] -> { version: version, release: release, image_src: image_src }
    mirroring = {}
    for i, image in enumerate(payload_images):
        latest_build = latest_builds[i]
        archives = archives_list[i]
        error = None
        if image.distgit_key in mismatched_siblings:
            error = "Siblings built from different commits"
        elif not (latest_build and archives):  # build or archive doesn't exist
            error = f"Unable to find build for: {image.image_name_short}"
            no_build_items.append(image.image_name_short)
        else:
            for archive in archives:
                arch = archive["arch"]
                pullspecs = archive["extra"]["docker"]["repositories"]
                if not pullspecs or ":" not in pullspecs[-1]:  # in case of no pullspecs or invalid format
                    error = f"Unable to find pullspecs for: {image.image_name_short}"
                    red_print(error, file=sys.stderr)
                    state.record_image_fail(lstate, image, error, runtime.logger)
                    break
                # The tag that will be used in the imagestreams
                tag_name = image.image_name_short
                tag_name = tag_name[4:] if tag_name.startswith("ose-") else tag_name  # it _should_ but... to be safe
                digest = archive["extra"]['docker']['digests']['application/vnd.docker.distribution.manifest.v2+json']
                if not digest.startswith("sha256:"):  # It should start with sha256: for now. Let's raise an error if this changes.
                    raise ValueError(f"Received unrecognized digest {digest} for image {pullspecs[-1]}")
                mirroring_value = {'version': latest_build["version"], 'release': latest_build["release"], 'image_src': pullspecs[-1], 'digest': digest}
                embargoed = latest_build["id"] in embargoed_build_ids  # when public_upstreams are not configured, this is always false
                if not embargoed:  # exclude embargoed images from the ocp[-arch] imagestreams
                    runtime.logger.info(f"Adding {arch} image {pullspecs[-1]} to the public mirroring list with imagestream tag {tag_name}...")
                    mirroring.setdefault(arch, {})[tag_name] = mirroring_value
                else:
                    red_print(f"Found embargoed image {pullspecs[-1]}")
                if runtime.group_config.public_upstreams:
                    # when public_upstreams are configured, both embargoed and non-embargoed images should be included in the ocp[-arch]-priv imagestreams
                    runtime.logger.info(f"Adding {arch} image {pullspecs[-1]} to the private mirroring list with imagestream tag {tag_name}...")
                    mirroring.setdefault(f"{arch}-priv", {})[tag_name] = mirroring_value
        if not error:
            state.record_image_success(lstate, image)
        else:
            red_print(error, file=sys.stderr)
            state.record_image_fail(lstate, image, error, runtime.logger)

    for key in mirroring:
        private = key.endswith("-priv")
        arch = key[:-5] if private else key  # strip `-priv` suffix

        mirror_filename = 'src_dest.{}'.format(key)
        imagestream_filename = 'image_stream.{}'.format(key)
        target_is_name, target_is_namespace = is_name_and_space(is_name, is_namespace, arch, private)

        def build_dest_name(tag_name):
            entry = mirroring[key][tag_name]
            tag = entry["digest"].replace(":", "-")  # sha256:abcdef -> sha256-abcdef
            return f"quay.io/{orgrepo}:{tag}"

        # Save the default SRC=DEST 'oc image mirror' input to a file for
        # later.
        with io.open(mirror_filename, 'w+', encoding="utf-8") as out_file:
            for tag_name in mirroring[key]:
                dest = build_dest_name(tag_name)
                out_file.write("{}={}\n".format(mirroring[key][tag_name]['image_src'], dest))

        with io.open("{}.yaml".format(imagestream_filename), 'w+', encoding="utf-8") as out_file:
            # Add a tag spec to the image stream. The name of each tag
            # spec does not include the 'ose-' prefix. This keeps them
            # consistent between OKD and OCP

            # Template Base Image Stream object.
            tag_list = []
            isb = {
                'kind': 'ImageStream',
                'apiVersion': 'image.openshift.io/v1',
                'metadata': {
                    'name': target_is_name,
                    'namespace': target_is_namespace,
                },
                'spec': {
                    'tags': tag_list,
                }
            }

            for tag_name in mirroring[key]:
                tag_list.append({
                    'name': tag_name,
                    'from': {
                        'kind': 'DockerImage',
                        'name': build_dest_name(tag_name)
                    }
                })

            # mirroring rhcos
            runtime.logger.info(f"Getting latest RHCOS pullspec for {target_is_name}...")
            mosc_istag = _latest_mosc_istag(runtime, arch, private)
            if mosc_istag:
                tag_list.append(mosc_istag)

            # Not all images are built for non-x86 arches (e.g. kuryr), but they
            # may be mentioned in image references. Thus, make sure there is a tag
            # for every tag we find in x86_64 and provide just a dummy image.
            if 'cli' not in mirroring[key]:  # `cli` serves as the dummy image for the replacement
                if runtime.group_config.public_upstreams and not private:  # If cli is embargoed, it is expected that cli is missing in any non *-priv imagestreams.
                    runtime.logger.warning(f"Unable to find cli tag from {key} imagestream. Is `cli` image embargoed?")
                else:  # if CVE embargoes supporting is disabled or the "cli" image is also missing in *-priv namespaces, an error will be raised.
                    raise DoozerFatalError('A dummy image is required for tag {} on arch {}, but unable to find cli tag for this arch'.format(tag_name, arch))
            else:
                extra_tags = mirroring['x86_64-priv' if private else 'x86_64'].keys() - mirroring[key].keys()
                for tag_name in extra_tags:
                    yellow_print('Unable to find tag {} for arch {} ; substituting cli image'.format(tag_name, arch))
                    tag_list.append({
                        'name': tag_name,
                        'from': {
                            'kind': 'DockerImage',
                            'name': build_dest_name('cli')  # cli is always built and is harmless
                        }
                    })

            yaml.safe_dump(isb, out_file, indent=2, default_flow_style=False)

    if no_build_items:
        yellow_print("No builds found for:")
        for img in sorted(no_build_items):
            click.echo("   {}".format(img))

    if invalid_name_items:
        yellow_print("Images skipped due to invalid naming:")
        for img in sorted(invalid_name_items):
            click.echo("   {}".format(img))

    if mismatched_siblings:
        yellow_print("Images skipped due to siblings mismatch:")
        for img in sorted(invalid_name_items):
            click.echo("   {}".format(img))


def default_is_base_name(version):
    return f"{version}-art-latest"


def default_is_base_namespace():
    return "ocp"


def is_name_and_space(base_name, base_namespace, arch, private):
    arch_suffix = "" if arch == 'x86_64' else f"-{arch}"
    priv_suffix = "-priv" if private else ""
    name = f"{base_name}{arch_suffix}{priv_suffix}"
    namespace = f"{base_namespace}{arch_suffix}{priv_suffix}"
    return name, namespace


def _latest_mosc_istag(runtime, arch, private):
    try:
        version = runtime.get_minor_version()
        _, pullspec = rhcos.latest_machine_os_content(version, arch, private)
        if not pullspec:
            yellow_print(f"No RHCOS found for {version} arch={arch} private={private}")
            return None
    except Exception as ex:
        yellow_print(f"error finding RHCOS: {ex}")
        return None

    return {
        'name': "machine-os-content",
        'from': {
            'kind': 'DockerImage',
            'name': pullspec
        }
    }


def find_mismatched_siblings(images, builds, archives_list, logger, lstate):
    """ Sibling images are those built from the same repository. We need to throw an error if there are sibling built from different commit.
    """
    # First, loop over all builds and store their source repos and commits to a dict
    repo_commit_nvrs = {}  # key is source repo url, value is another dict that key is commit hash and value is a set of nvrs.
    for build, archives in zip(builds, archives_list):
        if not build or not archives:
            continue  # the image has no builds yet. should be captured by latter logic.
        # source repo url and commit hash are stored in image's environment variables.
        ar = archives[0]  # in case the build is a manifest list, let's look at the first architecture
        envs = ar["extra"]["docker"]["config"]["config"].get("Env", [])
        source_repo_entry = list(filter(lambda env: env.startswith("SOURCE_GIT_URL="), envs))
        source_commit_entry = list(filter(lambda env: env.startswith("SOURCE_GIT_COMMIT="), envs))
        if not source_repo_entry or not source_commit_entry:
            continue  # this image doesn't have required environment variables. is it a dist-git only image?
        source_repo = source_repo_entry[0][source_repo_entry[0].find("=") + 1:]  # SOURCE_GIT_URL=https://example.com => https://example.com
        source_commit = source_commit_entry[0][source_commit_entry[0].find("=") + 1:]  # SOURCE_GIT_COMMIT=abc => abc
        nvrs = repo_commit_nvrs.setdefault(source_repo, {}).setdefault(source_commit, set())
        nvrs.add(build["nvr"])

    # Second, build a dict with keys are NVRs and values are the ImageMetadata objects. ImageMetadatas are used for logging state.
    nvr_images = {}
    for image, build in zip(images, builds):
        if build:
            nvr_images[build["nvr"]] = image

    # Finally, look at the dict and print an error if one repo has 2 or more commits
    mismatched_siblings = set()
    for repo, commit_nvrs in repo_commit_nvrs.items():
        if len(commit_nvrs) >= 2:
            red_print("The following NVRs are siblings but built from different commits:", file=sys.stderr)
            for commit, nvrs in commit_nvrs.items():
                for nvr in nvrs:
                    image = nvr_images[nvr]
                    mismatched_siblings.add(image.distgit_key)
                    red_print(f"{nvr}\t{image.distgit_key}\t{repo}\t{commit}")
    return mismatched_siblings



from __future__ import absolute_import, print_function, unicode_literals
import os
import io
from future.standard_library import install_aliases
install_aliases()
from urllib.parse import urlparse
import requests

from doozerlib.logutil import getLogger
from doozerlib.util import mkdirs, is_in_directory
from doozerlib.model import Missing
from doozerlib.exceptions import DoozerFatalError
from doozerlib.exectools import cmd_assert
from doozerlib.pushd import Dir

LOGGER = getLogger(__name__)


class SourceModifierFactory(object):
    """A factory class for creating source modifier objects."""
    MODIFICATIONS = {}

    @classmethod
    def supports(cls, action_name):
        """Test if specified modification action is supported"""
        return action_name in cls.MODIFICATIONS

    def create(self, *args, **kwargs):
        """Create a source modifier based on action.

        For example, create a source modifier for adding an out-of-tree file:

            factory = SourceModifierFactory()
            modifier = factory.create(action='add', source='http://example.com/gating_yaml', dest='gating.yaml', overwrite=True)
            modifier.modify()
        """
        action = kwargs["action"]
        if not self.supports(action):
            raise KeyError("Unknown modification action: {}.".format(action))
        return self.MODIFICATIONS[action](*args, **kwargs)


class AddModifier(object):
    """ A source modifier that supports adding an out-of-tree source to dist-git.

    An `add` action has the following valid fields:

    - `action`: must be `add`
    - `source`: URL to the out-of-tree source
    - `path`: Path in dist-git to write the source to
    - `overwriting`: Allow to overwrite if `path` exists

    For example, to add an out-of-tree source https://gitlab.cee.redhat.com/aosqe/ocp-build-data-gating/raw/master/openshift-3.11/atomic-openshift-cluster-autoscaler/gating_yaml to dist-git and save as `gating.yaml`:

        content:
          source:
            git:
              branch:
                fallback: master
                target: release-{MAJOR}.{MINOR}
              url: git@github.com:openshift/kubernetes-autoscaler.git
            modifications:
            - action: replace
              match: origin-cluster-autoscaler
              replacement: atomic-openshift-cluster-autoscaler
            - action: add
              source: https://gitlab.cee.redhat.com/aosqe/ocp-build-data-gating/raw/master/openshift-3.11/atomic-openshift-cluster-autoscaler/gating_yaml
              path: gating.yaml
              overwriting: true
            path: images/cluster-autoscaler
        # omitted
    """

    SUPPORTED_URL_SCHEMES = ["http", "https"]

    def __init__(self, *args, **kwargs):
        """ Initialize an "add" Modifier.
        :param source: URL to the out-of-tree source.
        :param path: Destination path to the dist-git repo.
        :param overwriting: True to allow to overwrite if path exists.
          Setting to false to prevent from accidently overwriting files from in-tree source.

        """
        self.source = kwargs["source"]
        self.path = kwargs["path"]
        self.overwriting = kwargs.get("overwriting", False)

    def act(self, *args, **kwargs):
        """ Run the modification action

        :param ceiling_dir: If not None, prevent from writing to a directory that is out of ceiling_dir.
        :param session: If not None, a requests.Session object for HTTP requests
        """
        LOGGER.debug("Running 'add' modification action...")
        context = kwargs["context"]
        distgit_path = context['distgit_path']
        source = urlparse(self.source)
        if source.scheme not in self.SUPPORTED_URL_SCHEMES:
            raise ValueError(
                "Unsupported URL scheme {} used in 'add' action.".format(source.scheme))
        source_url = source.geturl()  # normalized URL
        path = str(distgit_path.joinpath(self.path))
        ceiling_dir = kwargs.get("ceiling_dir")
        session = kwargs.get("session") or requests.session()
        if ceiling_dir and not is_in_directory(path, ceiling_dir):
            raise ValueError("Writing to a file out of {} is not allowed.".format(ceiling_dir))
        # NOTE: `overwriting` is checked before writing.
        # Data race might happen but it should suffice for prevent from accidently overwriting in-tree sources.
        if not self.overwriting and os.path.exists(path):
            raise IOError(
                "Destination path {} exists. Use 'overwriting: true' to overwrite.".format(self.path))
        LOGGER.debug("Getting out-of-tree source {}...".format(source_url))
        response = session.get(source_url)
        response.raise_for_status()
        mkdirs(os.path.dirname(path))
        with io.open(path, "wb") as dest_file:
            dest_file.write(response.content)
        LOGGER.debug("Out-of-tree source saved: {} -> {}".format(source_url, path))


SourceModifierFactory.MODIFICATIONS["add"] = AddModifier


class ReplaceModifier(object):
    """ A source modifier that supports replacing a substring in Dockerfile or RPM spec file.
    """

    def __init__(self, *args, **kwargs):
        """ Initialize ReplaceModifier
        :param match: This is old substring to be replaced.
        :param replacement: This is new substring, which would replace old substring.
        """
        self.match = kwargs["match"]
        self.replacement = kwargs["replacement"]

    def act(self, *args, **kwargs):
        """ Run the modification action

        :param context: A context dict. `context.component_name` is the dist-git repo name,
            and `context.content` is the content of Dockerfile or RPM spec file.
        """
        context = kwargs["context"]
        content = context["content"]
        component_name = context["component_name"]
        match = self.match
        assert (match is not Missing)
        replacement = self.replacement
        assert (replacement is not Missing)
        if replacement is None:  # Nothing follows colon in config yaml; user attempting to remove string
            replacement = ""
        pre = content
        post = pre.replace(match, replacement)
        if post == pre:
            raise DoozerFatalError("{}: Replace ({}->{}) modification did not make a change to the Dockerfile content"
                                   .format(component_name, match, replacement))
        LOGGER.debug(
            "Performed string replace '%s' -> '%s':\n%s\n" %
            (match, replacement, post))
        context["content"] = post


SourceModifierFactory.MODIFICATIONS["replace"] = ReplaceModifier


class CommandModifier(object):
    """ A source modifier that supports running a custom command to modify the source.
    """

    def __init__(self, *args, **kwargs):
        """ Initialize CommandModifier
        :param command: a `str` or `list` of the command with arguments
        """
        self.command = kwargs["command"]

    def act(self, *args, **kwargs):
        """ Run the command
        :param context: A context dict. `context.set_env` is a `dict` of env vars to set for command (overriding existing).
        """
        context = kwargs["context"]
        set_env = context["set_env"]
        with Dir(context['distgit_path']):
            cmd_assert(self.command, set_env=set_env)


SourceModifierFactory.MODIFICATIONS["command"] = CommandModifier
